import numpy as np
from scipy.optimize import leastsq


def fitfunc(p, x):
    return p[0] + p[1] * (x ** p[2])


def errfunc(p, x, y):
    return np.power(y - fitfunc(p, x), 2)


class PowerLaw(object):
    def __init__(self):
        self.partial_func = None

    def fit(self, xdata, ydata):
        self.params, _ = leastsq(errfunc, [max(ydata), -1, -0.5], args=(xdata, ydata), maxfev=500)
        return self

    def predict(self, x):
        return fitfunc(self.params, x)



import numpy as np
import pandas as pd
from scipy.stats import rankdata

from scattertext import CorpusBasedTermScorer


class RankDifferenceScorer(CorpusBasedTermScorer):
    def _set_scorer_args(self, **kwargs):
        '''
        method : {'average', 'min', 'max', 'dense', 'ordinal'}, optional
        '''
        self.method = kwargs.get('method', 'dense')

    def get_scores(self, *args):
        '''
        In this case, args aren't used, since this information is taken
        directly from the corpus categories.

        Returns
        -------
        np.array, scores
        '''
        if self.tdf_ is None:
            raise Exception("Use set_categories('category name', ['not category name', ...]) " +
                            "to set the category of interest")

        dense_rank_difference = self._dense_rank_difference(
            self.tdf_['cat'], self.tdf_['ncat']
        )

        if type(self.tdf_['cat']) == pd.Series:
            return pd.Series(
                dense_rank_difference,
                index=self.tdf_['cat'].index
            )
        return dense_rank_difference

    def _dense_rank_difference(self, a, b):
        a_ranks = rankdata(a, self.method)
        b_ranks = rankdata(b, self.method)
        to_ret = (a_ranks / np.max(a_ranks)) - (b_ranks / np.max(b_ranks))
        return to_ret

    def get_default_score(self):
        return 0

    def get_name(self):
        return 'Rank Difference'



import json
from bisect import bisect_left

import pandas as pd
import numpy as np
from scattertext.termranking import AbsoluteFrequencyRanker

from scattertext.termscoring.RankDifference import RankDifference
from scipy.stats import gmean
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.linear_model import LogisticRegression

from scattertext import Scalers
from scattertext.graphs.ComponentDiGraphHTMLRenderer import GraphRenderer
from scattertext.viz.HTMLSemioticSquareViz import ClickableTerms


class MultiCategoryAssociationBase:
    def __init__(self, corpus, use_metadata=False):
        self.corpus = corpus
        self.use_metadata = use_metadata

    def get_category_association(self):
        raise NotImplementedError()


class MultiCategoryAssociationScorer(MultiCategoryAssociationBase):
    def get_category_association(self, ranker=None, scorer=None):
        if scorer is None:
            scorer = RankDifference()
        if ranker is None:
            ranker = AbsoluteFrequencyRanker(self.corpus)
        if self.use_metadata:
            ranker = ranker.use_non_text_features()
        term_freq_df = ranker.get_ranks('')
        global_freq = term_freq_df.sum(axis=1)
        data = []
        for cat in self.corpus.get_categories():
            cat_freq = term_freq_df[cat]
            for term_rank, (term, score) in enumerate(scorer.get_scores(
                    cat_freq, global_freq - cat_freq
            ).sort_values(ascending=False).iteritems()):
                data.append({'Category': cat, 'Term': term, 'Rank': term_rank, 'Score': score})

        return pd.DataFrame(data).groupby('Rank')


class CategoryTableMaker(GraphRenderer):
    def __init__(
            self,
            corpus,
            num_rows=10,
            use_metadata=False,
            category_order=None,
    ):
        self.num_rows = num_rows
        self.corpus = corpus
        self.use_metadata = use_metadata
        self.rank_df = self._get_term_category_associations()
        self.category_order = (sorted(self.corpus.get_categories())
                               if category_order is None else category_order)

    def get_graph(self):

        table = '<div class="timelinecontainer"><table class="timelinetable">'
        table += '<tr><th>' + '</th><th>'.join(self.category_order) + '</th></tr>'
        min_font_size = 7
        max_font_size = 20
        display_df = self.rank_df[lambda df: df.Rank < self.num_rows]

        bin_boundaries = np.histogram_bin_edges(
            np.log(display_df.Frequency), bins=max_font_size - min_font_size
        )
        display_df = pd.merge(
            display_df.assign(
                FontSize=lambda df: df.Frequency.apply(np.log).apply(
                    lambda x: bisect_left(bin_boundaries, x) + min_font_size
                )
            ),
            pd.DataFrame({'Category': self.category_order, 'CategoryNum': np.arange(len(self.category_order))}),
            on='Category'
        )

        for rank, group_df in display_df.groupby('Rank'):
            table += '<tr><td class="clickabletd">' + '</td><td class="clickabletd">'.join([
                ClickableTerms.get_clickable_term(
                    row.Term,
                    style="font-size: " + str(row.FontSize)
                )
                for _, row in group_df.sort_values(by='CategoryNum').iterrows()
            ]) + '</td></tr>'
        table += '</table></div>'
        return table

    def get_javascript(self):
        d = {}
        for category, cat_df in self.rank_df.assign(Frequency=lambda df: df.Frequency.astype(int)).groupby('Category'):
            cat_d = {}
            for _, row in cat_df.iterrows():
                cat_d[row['Term']] = {'Rank': row['Rank'], 'Freq': row['Frequency']}
            d[category] = cat_d
        js = 'categoryFrequency = ' + json.dumps(d) + ';\n\n'

        js += '''
        Array.from(document.querySelectorAll('.clickabletd')).map(
            function (node) {
                node.addEventListener('mouseenter', mouseEnterNode);
                node.addEventListener('mouseleave', mouseLeaveNode);
                node.addEventListener('click', clickNode);
            }
        )
        
        function clickNode() {
            //document.querySelectorAll(".dotgraph")
            //    .forEach(node => node.style.display = 'none');

            //var term = Array.prototype.filter
            //    .call(this.children, (x => x.tagName === "span"))[0].textContent;

            //plotInterface.handleSearchTerm(term, true);
            
            
        }

        function mouseEnterNode(event) {
            console.log("THIS"); console.log(this)
            var term = this.children[0].textContent;
            plotInterface.showTooltipSimple(term);
            var clickableTds = document.getElementsByClassName('clickabletd');
            for(var i = 0; i < clickableTds.length; i++) {
                var td = clickableTds[i];
                if(td.children[0].textContent == term) 
                    td.style.backgroundColor = "#FFAAAA";
            }
        }

        function mouseLeaveNode(event) {
            plotInterface.tooltip.transition().style('opacity', 0)
            var clickableTds = document.getElementsByClassName('clickabletd');            
            for(var i = 0; i < clickableTds.length; i++) 
                clickableTds[i].style.backgroundColor = "#FFFFFF";
        }'''
        return js

    def _get_term_category_associations(self):
        tdm = self.corpus.get_metadata_doc_mat() if self.use_metadata else self.corpus.get_term_doc_mat()
        tdmtfidf = TfidfTransformer().fit_transform(tdm)
        coefs = np.zeros(shape=(self.corpus.get_num_categories(), tdm.shape[1]), dtype=float)
        for i, cat in enumerate(self.corpus.get_categories()):
            y = self.corpus.get_category_ids() == i
            clf = LogisticRegression(
                penalty='l2', C=5., max_iter=4000, tol=1e-6, solver='liblinear'
            ).fit(tdmtfidf, y)

            coefs[i, :] = clf.coef_
        coef_df = pd.DataFrame(
            coefs.T,
            index=self.corpus.get_terms(use_metadata=self.use_metadata),
            columns=[x + ' coef' for x in self.corpus.get_categories()]
        )
        coef_freq_df = pd.merge(
            self.corpus.get_freq_df(self.use_metadata),
            coef_df,
            left_index=True,
            right_index=True
        )
        for cat in self.corpus.get_categories():
            freq = Scalers.dense_rank(coef_freq_df[cat + ' freq'])
            coef = Scalers.scale_neg_1_to_1_with_zero_mean(coef_freq_df[cat + ' coef'])
            coef_freq_df[cat + ' gmean'] = gmean([freq, coef, coef])
        data = []
        for cat in self.corpus.get_categories():
            for term_rank, (term, row) in enumerate(
                    coef_freq_df[[
                        cat + ' gmean', cat + ' freq'
                    ]].sort_values(by=cat + ' gmean', ascending=False).iterrows()
            ):
                data.append({'Category': cat, 'Term': term, 'Rank': term_rank, 'Frequency': row[cat + ' freq']})
        return pd.DataFrame(data)



import numpy as np
import pandas as pd

DEFAULT_GRAPH_PARAMS = {
    'charset': "latin1",
    'outputorder': "edgesfirst",
    'overlap': "prism"
}

DEFAULT_NODE_PARAMS = {
    'fontname': "'IBM Plex Sans'",
    'fontsize': 10
}


class ComponentDiGraph(object):
    def __init__(self, orig_edge_df, id_node_df, node_df, edge_df, components, graph_params=None, node_params=None):
        self.edge_df = edge_df
        self.orig_edge_df = orig_edge_df
        self.id_node_df = id_node_df
        self.node_df = node_df
        self.components = components
        self.graph_params = DEFAULT_GRAPH_PARAMS
        if graph_params is not None:
            for k, v in graph_params.items():
                self.graph_params[k] = v
        self.node_params = DEFAULT_NODE_PARAMS
        if node_params is not None:
            for k, v in node_params.items():
                self.node_params[k] = v

    def get_dot(self, component):
        # graph = '''digraph {\n node [fixedsize="true", fontname="'IBM Plex Sans'", height="0.0001", label="\n", margin="0", shape="plaintext", width="0.0001"];\n'''

        # graph = '''digraph  \n{\n graph [bb="0,0,1297.2,881.5", charset="latin1", outputorder="edgesfirst", overlap="prism"]\n node [fontname="'IBM Plex Sans'", fontsize=10] ;\n'''
        graph = '''digraph  \n{\n graph [%s]\n node [%s] ;\n''' % (
            self._format_graphviz_paramters(self.graph_params),
            self._format_graphviz_paramters(self.node_params)
        )

        mynode_df = self.id_node_df[self.id_node_df['Component'] == component]
        nodes = '\n'.join(mynode_df
                          .reset_index()
                          .apply(lambda x: '''"%s" [label="%s"] ;''' % (x['index'], x['name']), axis=1)
                          .values)
        edges = '\n'.join(self.edge_df[self.edge_df.Component == component].
                          apply(lambda x: '"%s" -> "%s" ;' % (x.source_id, x.target_id), axis=1).values)
        return graph + nodes + '\n\n' + edges + '\n}'

    def get_components_at_least_size(self, min_size):
        component_sizes = (pd.DataFrame({'component': self.components})
                           .reset_index()
                           .groupby('component')[['index']]
                           .apply(len)
                           .where(lambda x: x >= min_size).dropna())
        return np.array(component_sizes.sort_values(ascending=False).index)

    def get_node_to_component_dict(self):
        return self.id_node_df.set_index('name')['Component'].to_dict()

    def component_to_node_list_dict(self):
        return self.id_node_df.groupby('Component')['name'].apply(list).to_dict()

    def _format_graphviz_paramters(self, params):
        return ', '.join([k + '=' + ('"' if type(v) == str else '') + str(v) + ('"' if type(v) == str else '')
                          for k, v in params.items()])



from collections import Counter
from itertools import chain
import pandas as pd


class FeatsFromSpacyDoc(object):
	def __init__(self,
	             use_lemmas=False,
	             entity_types_to_censor=set(),
	             tag_types_to_censor=set(),
	             strip_final_period=False):
		'''
		Parameters
		----------
		use_lemmas : bool, optional
			False by default
		entity_types_to_censor : set, optional
			empty by default
		tag_types_to_censor : set, optional
			empty by default
		strip_final_period : bool, optional
			if you know that spacy is going to mess up parsing, strip final period.  default no.
		'''
		self._use_lemmas = use_lemmas
		assert type(entity_types_to_censor) == set
		assert type(tag_types_to_censor) == set
		self._entity_types_to_censor = entity_types_to_censor
		self._pos_types_to_censor = {}
		self._tag_types_to_censor = tag_types_to_censor
		self._strip_final_period = strip_final_period
		self._ignore_censored_types = False

	def _post_process_term(self, term):
		if self._strip_final_period and (term.strip().endswith('.') or term.strip().endswith(',')):
			term = term.strip()[:-1]
		return term

	def get_doc_metadata(self, doc):
		return Counter()

	def get_row_metadata(self, doc, row: pd.Series):
		return Counter()

	def ignore_censored_types(self):
		self._ignore_censored_types = True
		return self

	def censor_pos_types(self, pos_types):
		assert type(pos_types) == set
		self._pos_types_to_censor = pos_types
		return self

	def censor_entity_types(self, entity_types):
		assert type(entity_types) == set
		self._entity_types_to_censor = entity_types
		return self

	def censor_tag_types(self, tag_types):
		assert type(tag_types) == set
		self._tag_types_to_censor = tag_types
		return self

	def get_feats(self, doc):
		'''
		Parameters
		----------
		doc, Spacy Docs

		Returns
		-------
		Counter (unigram, bigram) -> count
		'''
		ngram_counter = Counter()
		for sent in doc.sents:
			unigrams = self._get_unigram_feats(sent)
			bigrams = self._get_bigram_feats(unigrams)
			ngram_counter += Counter(chain(unigrams, bigrams))
		return ngram_counter

	def _get_bigram_feats(self, unigrams):
		if len(unigrams) > 1:
			bigrams = map(' '.join, zip(unigrams[:-1], unigrams[1:]))
		else:
			bigrams = []
		return bigrams

	def _get_unigram_feats(self, sent):
		unigrams = []
		for tok in sent:
			if tok.pos_ not in ('PUNCT', 'SPACE', 'X'):
				if tok.ent_type_ in self._entity_types_to_censor:
					if not self._ignore_censored_types:
						unigrams.append('_' + tok.ent_type_)
				elif tok.tag_ in self._tag_types_to_censor:
					if not self._ignore_censored_types:
						unigrams.append(tok.tag_)
				elif tok.pos_ in self._pos_types_to_censor:
					if not self._ignore_censored_types:
						unigrams.append(tok.pos_)
				elif self._use_lemmas and tok.lemma_.strip():
					unigrams.append(self._post_process_term(tok.lemma_.strip().lower()))
				elif tok.lower_.strip():
					unigrams.append(self._post_process_term(tok.lower_.strip()))
		return unigrams

	def has_metadata_term_list(self):
		'''
		Returns True if there is a meta data term list associated with object, False if not.

		Returns
		-------
		bool
		'''
		return False

	def get_top_model_term_lists(self):
		raise Exception("No topic models associated with these features.")


import string

import numpy as np

from scattertext.OffsetCorpus import OffsetCorpus

from scattertext.CSRMatrixTools import CSRMatrixFactory
from scattertext.indexstore.IndexStore import IndexStore


class OffsetCorpusFactory(object):
    def __init__(self,
                 df,
                 parsed_col,
                 feat_and_offset_getter,
                 category_col=None):

        '''
        Parameters
        ----------
        df : pd.DataFrame
         contains category_col, and parse_col, were parsed col is entirely spacy docs
        parsed_col : str
            name of spacy parsed column in convention_df
        feats_from_spacy_doc : FeatsFromSpacyDoc
        category_col : str, Optional
            name of category column in df; if None, all category names will be '_'
        '''
        self._df = df.reset_index()
        self._category_col = category_col
        self._parsed_col = parsed_col
        self._category_idx_store = IndexStore()
        self._X_factory = CSRMatrixFactory()
        self._mX_factory = CSRMatrixFactory()
        self._term_idx_store = IndexStore()
        self._metadata_idx_store = IndexStore()
        self._feat_and_offset_getter = feat_and_offset_getter
        self._term_offsets = {}
        self._metadata_offsets = {}

    def build(self, show_progress=False):
        '''Constructs the term doc matrix.

        Returns
        -------
        scattertext.ParsedCorpus.ParsedCorpus
        '''
        self._ensure_category_col_is_in_df()
        y = self._get_y_and_populate_category_idx_store(self._df[self._category_col])
        if show_progress is True:
            self._df.progress_apply(self._add_to_x_factory, axis=1)
        else:
            self._df.apply(self._add_to_x_factory, axis=1)

        self._mX = self._mX_factory.set_last_row_idx(len(y) - 1).get_csr_matrix()
        return OffsetCorpus(
            df=self._df,
            X=self._X_factory.set_last_row_idx(len(y) - 1).get_csr_matrix(),
            mX=self._mX_factory.set_last_row_idx(len(y) - 1).get_csr_matrix(),
            y=y,
            term_idx_store=self._term_idx_store,
            category_idx_store=self._category_idx_store,
            metadata_idx_store=self._metadata_idx_store,
            parsed_col=self._parsed_col,
            category_col=self._category_col,
            term_offsets=self._term_offsets,
            metadata_offsets=self._metadata_offsets
        )

    def _ensure_category_col_is_in_df(self):
        if self._category_col not in self._df:
            self._category_col = 'Category'
            while self._category_col in self._df:
                self._category_col = 'Category_' + ''.join(np.random.choice(string.ascii_letters) for _ in range(5))
            self._df[self._category_col] = ''
    def _get_y_and_populate_category_idx_store(self, categories):
        return np.array(categories.apply(self._category_idx_store.getidx))

    def _add_to_x_factory(self, row):
        parsed_text = row[self._parsed_col]
        for term, (count, offsets) in self._feat_and_offset_getter.get_term_offsets(parsed_text):
            term_idx = self._term_idx_store.getidx(term)
            self._X_factory[row.name, term_idx] = count
            if offsets is not None:
                self._term_offsets.setdefault(term, {}).setdefault(row.name, []).extend(offsets)

        for meta, (val, offsets) in self._feat_and_offset_getter.get_metadata_offsets(parsed_text):
            self.__get_metadata_offsets(meta, offsets, row, val)

        for meta, (val, offsets) in self._feat_and_offset_getter.get_metadata_row_offsets(parsed_text, row):
            self.__get_metadata_offsets(meta, offsets, row, val)

    def __get_metadata_offsets(self, meta, offsets, row, val):
        meta_idx = self._metadata_idx_store.getidx(meta)
        self._mX_factory[row.name, meta_idx] = val
        if offsets is not None:
            self._metadata_offsets.setdefault(meta, {}).setdefault(row.name, []).extend(offsets)



class FeatAndOffsetGetter(object):
    def get_term_offsets(self, doc):
        return None

    def get_metadata_offsets(self, doc):
        return None

    def get_metadata_row_offsets(self, doc, row):
        return []


import numpy as np
import pandas as pd

from scattertext.termscoring.CorpusBasedTermScorer import CorpusBasedTermScorer


class CraigsZeta(CorpusBasedTermScorer):
    """
    Zeta function from:

    Pervez Rizvi. The interpretation of Zeta test results.
    Digital Scholarship in the Humanities, Volume 34, Issue 2, June 2019, Pages 401–418,

    The score assumes that each document ("segment" in this literature) is approximately of equal size.
    Rizvi reports segment sizes being 900-6,000 words.

    Z_t = (# segments with type t labeled in the target category)/(# segments in target category) -
     (# segments with type t labeled not in the target category)/(# segments not in target category)

    """

    def _set_scorer_args(self, **kwargs):
        self.constant = kwargs.get('constant', 0.01)

    def get_scores(self, *args):
        '''
        In this case, args aren't used, since this information is taken
        directly from the corpus categories.

        Returns
        -------
        np.array, scores
        '''
        X = self._get_X() > 0
        cat_X, ncat_X = self._get_cat_and_ncat(X)
        return self._get_zeta_score(cat_X, ncat_X)

    def get_score_df(self, *args):
        '''
        In this case, args aren't used, since this information is taken
        directly from the corpus categories.

        Returns
        -------
        np.array, scores
        '''
        X = self._get_X() > 0
        cat_X, ncat_X = self._get_cat_and_ncat(X)
        score = self._get_zeta_score(cat_X, ncat_X)
        return pd.DataFrame({
            'Term': self.corpus_.get_terms(),
            'Base': cat_X.sum(axis=0).A1,
            'Counter': ncat_X.sum(axis=0).A1,
            'Score': score
        }).set_index('Term')

    def _get_zeta_score(self, cat_X, ncat_X):
        return (
                (cat_X.sum(axis=0) + self.constant) / (cat_X.shape[0] + self.constant)
                - (ncat_X.sum(axis=0) + self.constant) / (ncat_X.shape[0] + self.constant)
        ).A1

    def get_name(self):
        return 'Craigs Zeta'

class ChiSquareZeta(CraigsZeta):
    def _get_zeta_score(self, cat_X, ncat_X):
        return (
                (cat_X.sum(axis=0) + self.constant) / (cat_X.shape[0] + self.constant)
                - (ncat_X.sum(axis=0) + self.constant) / (ncat_X.shape[0] + self.constant)
        ).A1


import numpy as np

from scattertext.termscoring.CorpusBasedTermScorer import CorpusBasedTermScorer


class LogOddsRatio(CorpusBasedTermScorer):

    def _set_scorer_args(self, **kwargs):
        self.constant = kwargs.get('constant', 0.01)

    def get_scores(self, *args):
        '''
        In this case, args aren't used, since this information is taken
        directly from the corpus categories.

        Returns
        -------
        np.array, scores
        '''
        if self.tdf_ is None:
            raise Exception("Use set_categories('category name', ['not category name', ...]) " +
                            "to set the category of interest")

        y_i = self.tdf_['cat']
        y_j = self.tdf_['ncat']
        n_i, n_j = y_i.sum(), y_j.sum()
        delta_i_j = (np.log((y_i + self.constant) / (self.constant + n_i - y_i))
                     - np.log((y_j + self.constant) / (self.constant + n_j - y_j)))
        return delta_i_j

    def get_name(self):
        return 'Smoothed Log Odds Ratio'




import pandas as pd
import numpy as np
from sklearn.isotonic import IsotonicRegression


class MeanIsotonic:
    def __init__(self, n=1000):
        self.n = n

    def fit_predict(self, x, y):
        assert len(x) == len(y)
        df = pd.DataFrame({
            'x': x,
            'y': y
        })

        # Average runs of isotonic regression
        pred = np.zeros(len(df), dtype=np.float)
        for i in range(self.n):
            sample_df = df.sample(frac=0.5)
            pred += 1 / self.n * IsotonicRegression(
                y_max=1, y_min=0, out_of_bounds='clip'
            ).fit(sample_df.x.values, sample_df.y.values).predict(df.x.values)
        return pred




from collections import Counter
from functools import partial
from sys import version_info

from scattertext.features.FeatsFromSpacyDoc import FeatsFromSpacyDoc


class FeatsFromSpacyDocAndEmpath(FeatsFromSpacyDoc):
    def __init__(self,
                 use_lemmas=False,
                 entity_types_to_censor=set(),
                 tag_types_to_censor=set(),
                 strip_final_period=False,
                 empath_analyze_function=None,
                 **kwargs):
        '''
        Parameters
        ----------
        empath_analyze_function: function (default=empath.Empath().analyze)
            Function that produces a dictionary mapping Empath categories to

        Other parameters from FeatsFromSpacyDoc.__init__
        '''
        if empath_analyze_function is None:
            try:
                import empath
            except ImportError:
                raise Exception("Please install the empath library to use FeatsFromSpacyDocAndEmpath.")
            self._empath_analyze_function = empath.Empath().analyze
        else:
            self._empath_analyze_function = partial(empath_analyze_function,
                                                    kwargs={'tokenizer': 'bigram'})
        FeatsFromSpacyDoc.__init__(self, use_lemmas,
                                   entity_types_to_censor,
                                   tag_types_to_censor,
                                   strip_final_period)

    def get_doc_metadata(self, doc, prefix=''):
        empath_counter = Counter()
        if version_info[0] >= 3:
            doc = str(doc)
        for empath_category, score in self._empath_analyze_function(doc).items():
            if score > 0:
                empath_counter[prefix + empath_category] = int(score)
        return empath_counter

    def has_metadata_term_list(self):
        return True

    def get_top_model_term_lists(self):
        try:
            import empath
        except ImportError:
            raise Exception("Please install the empath library to use FeatsFromSpacyDocAndEmpath.")
        return dict(empath.Empath().cats)



import itertools
import warnings
from collections import Counter

from scattertext.ParsedCorpus import ParsedCorpus


class FeatsFromGensim(object):
    def __init__(self, phrases, gram_size):
        '''
        Parameters
        ----------
        phrases : list[gensim.models.Phrases]
        gram_size : int, maximum number of words per phrase
        kwargs : parameters for FeatsFromSpacyDoc.init
        '''
        from gensim.models import Phrases

        phrases = phrases
        gram_size = gram_size
        assert type(phrases) == Phrases
        self.gram_size = gram_size
        self.phrases = phrases

    def get_doc_metadata(self, doc):
        return Counter()

    def get_feats(self, doc):
        '''
        Parameters
        ----------
        doc, Spacy Docs

        Returns
        -------
        Counter (unigram, bigram) -> count
        '''
        ngram_counter = Counter()
        for sent in doc.sents:
            ngrams = self.phrases[str(sent)]
            for subphrases in self.phrases[1:]:
                ngrams = subphrases[str(sent)]
            for ngram in ngrams:
                ngram_counter[ngram] += 1
        return ngram_counter


class GensimPhraseAdder(object):
    def __init__(self, max_tokens_per_phrase=3, phrases=None):
        '''
        Parameters
        ----------
        max_tokens_per_phrase: int, must be > 1.  Default 3
        phrases: Instance of Gensim phrases class, default None
        '''
        self.max_tokens_per_phrase = max_tokens_per_phrase
        self.phrases = phrases

    def add_phrases(self, corpus):
        '''
        Parameters
        ----------
        corpus: Corpus for phrase augmentation

        Returns
        -------
        New ParsedCorpus containing unigrams in corpus and new phrases
        '''
        from gensim.models import Phrases

        assert isinstance(corpus, ParsedCorpus)
        self.phrases = [Phrases(CorpusAdapterForGensim.get_sentences(corpus), delimiter=' ')]

        for i in range(1, self.max_tokens_per_phrase):
            self.phrases.append(Phrases(self.phrases[-1][CorpusAdapterForGensim.get_sentences(corpus)]))

        return self


class CorpusAdapterForGensim(object):
    @staticmethod
    def get_token_format(token):
        return token.lower_

    @classmethod
    def get_sentences(cls, corpus):
        '''
        Parameters
        ----------
        corpus, ParsedCorpus

        Returns
        -------
        iter: [sentence1word1, ...], [sentence2word1, ...]
        '''
        # assert isinstance(corpus, ParsedCorpus)
        return itertools.chain(*[[[cls.get_token_format(t)
                                   for t in sent
                                   if not t.is_punct]
                                  for sent in doc.sents]
                                 for doc in corpus.get_parsed_docs()])


class Word2VecDefault(object):
    def _default_word2vec_model(self):
        from gensim.models import word2vec
        return word2vec.Word2Vec(vector_size=100,
                                 alpha=0.025,
                                 window=5,
                                 min_count=5,
                                 max_vocab_size=None,
                                 sample=0,
                                 seed=1,
                                 workers=1,
                                 min_alpha=0.0001,
                                 sg=1,
                                 hs=1,
                                 negative=0,
                                 cbow_mean=0,
                                 # iter=1,
                                 null_word=0,
                                 trim_rule=None,
                                 sorted_vocab=1)


class Word2VecFromParsedCorpus(Word2VecDefault):
    def __init__(self, corpus, word2vec_model=None):
        '''
        Parameters
        ----------
        corpus: ParsedCorpus
          from which to build word2vec model
        word2vec_model: word2vec.Word2Vec
            Gensim instance to be used to train word2vec model
        '''
        try:
            from gensim.models import word2vec
            assert word2vec_model is None or isinstance(word2vec_model, word2vec.Word2Vec)
        except:
            warnings.warn("You should really install gensim, but we're going to duck-type your model and pray it works")
        # assert isinstance(corpus, ParsedCorpus)
        self.corpus = corpus
        self.model = self._get_word2vec_model(word2vec_model)

    def train(self, epochs=2000, super_epochs=5, tqdm=None):
        '''
        Parameters
        ----------
        epochs : int
          Number of epochs to train for.  Default is 2000.
        super_epochs : int
            Number of times to repeat training process. Default is training_iterations.

        Returns
        -------
        A trained word2vec model.
        '''

        self._scan_and_build_vocab()
        myiter = range(super_epochs)
        if tqdm:
            myiter = tqdm(myiter, total=super_epochs)
        for _ in myiter:
            self.model.train(CorpusAdapterForGensim.get_sentences(self.corpus),
                             total_examples=self.model.corpus_count,
                             epochs=epochs)
        return self.model

    def _get_word2vec_model(self, word2vec_model):
        return (self._default_word2vec_model()
                if word2vec_model is None
                else word2vec_model)

    def _scan_and_build_vocab(self):
        try:
            self.model.scan_vocab(CorpusAdapterForGensim.get_sentences(self.corpus))
        except:
            pass
        self.model.build_vocab(CorpusAdapterForGensim.get_sentences(self.corpus))


class Word2VecFromParsedCorpusBigrams(Word2VecFromParsedCorpus):
    def _scan_and_build_vocab(self):
        from gensim.models import Phrases
        bigram_transformer = Phrases(CorpusAdapterForGensim.get_sentences(self.corpus))
        try:
            self.model.scan_vocab(CorpusAdapterForGensim.get_sentences(self.corpus))
        except:
            pass
        self.model.build_vocab(bigram_transformer[CorpusAdapterForGensim.get_sentences(self.corpus)])



import pkgutil
from io import StringIO

import pandas as pd
from scipy.stats import rankdata


class BackgroundFrequencyDataFramePreparer(object):
    @staticmethod
    def prep_background_frequency(df):
        df['rank'] = rankdata(df.background, method='dense')
        df['background'] = df['rank'] / df['rank'].max()
        return df['background']


class BackgroundFrequenciesFromCorpus(BackgroundFrequencyDataFramePreparer):
    def __init__(self, corpus, exclude_categories=[]):
        self.background_df = (pd.DataFrame(corpus.remove_categories(exclude_categories)
                                           .get_term_freq_df().sum(axis=1))
                              .rename(columns={0: 'background'}))

    def get_background_frequency_df(self):
        return self.background_df

    def get_background_rank_df(self):
        return self.prep_background_frequency(self.get_background_frequency_df())


class BackgroundFrequencies(BackgroundFrequencyDataFramePreparer):
    @staticmethod
    def get_background_frequency_df(frequency_path=None):
        raise Exception

    @classmethod
    def get_background_rank_df(cls, frequency_path=None):
        return cls.prep_background_frequency(
            cls.get_background_frequency_df(frequency_path)
        )


class DefaultBackgroundFrequencies(BackgroundFrequencies):
    @staticmethod
    def get_background_frequency_df(frequency_path=None):
        if frequency_path:
            unigram_freq_table_buf = open(frequency_path)
        else:
            unigram_freq_table_buf = StringIO(pkgutil.get_data('scattertext', 'data/count_1w.txt')
                                              .decode('utf-8'))
        to_ret = (pd.read_csv(unigram_freq_table_buf,
                              sep='\t',
                              names=['word', 'background'])
                  .sort_values(ascending=False, by='background')
                  .drop_duplicates(['word'])
                  .set_index('word'))['background']
        return to_ret



import logging

import numpy as np
import pandas as pd
"""
seed_mat = np.matrix([model.wv[word] for word in pos_seeds + neg_seeds])
term_mat = np.matrix([model.wv[word] for word in model.wv.vocab])
seed_norm = np.matrix(np.linalg.norm(seed_mat, axis=1))
term_norm = np.matrix(np.linalg.norm(term_mat, axis=1))
cosine_mat = (((seed_mat * term_mat.T).T/seed_norm)/term_norm.T)
polarity_mat = np.matrix([1/len(pos_seeds)] * len(pos_seeds) + [-1/len(neg_seeds)] * len(neg_seeds))
term_scale = (cosine_mat * polarity_mat.T).A1
"""

def latent_semantic_scale_from_word2vec(model,
                                        pos_seed_words=None,
                                        neg_seed_words=None,
                                        seed_words=None,
                                        seed_values=None,):
    terms = [word for word in model.wv.key_to_index.keys()]
    embeddings = np.matrix([model.wv[word] for word in model.wv.key_to_index.keys()])
    return lss_terms(embeddings, terms, pos_seed_words, neg_seed_words, seed_words, seed_values)


def lss_terms(embeddings,
              terms,
              pos_seed_words=None,
              neg_seed_words=None,
              seed_words=None,
              seed_values=None):
    neg_seed_words = [] if neg_seed_words is None else neg_seed_words
    pos_seed_words = [] if pos_seed_words is None else pos_seed_words
    seed_words = [] if seed_words is None else seed_words
    seed_values = [] if seed_values is None else seed_values
    for word, value in zip((pos_seed_words + neg_seed_words), [1] * len(pos_seed_words) + [-1] * len(neg_seed_words)):
        seed_words.append(word)
        seed_values.append(value)

    assert seed_values
    assert embeddings.shape[0] == len(terms)
    assert len(seed_words) == len(seed_values)
    missing_words = []
    for word in seed_words:
        if word not in terms:
            missing_words.append(word)
    if missing_words:
        raise Exception(f"No embedding(s) exists for {','.join(missing_words)}.")

    term2i = {term:i for i, term in enumerate(terms)}
    seed_mat = embeddings[[term2i[term] for term in seed_words],:]

    seed_norm = np.matrix(np.linalg.norm(seed_mat, axis=1))
    term_norm = np.matrix(np.linalg.norm(embeddings, axis=1))

    cosine_mat = (((seed_mat * embeddings.T).T/seed_norm)/term_norm.T)

    return pd.Series((cosine_mat * np.matrix(seed_values).T).A1, index=terms)




from .ComponentDiGraph import ComponentDiGraph
from .ComponentDiGraphHTMLRenderer import ComponentDiGraphHTMLRenderer
from .GraphStructure import GraphStructure
from .SimpleDiGraph import SimpleDiGraph



import json

class GraphRenderer(object):
    def get_graph(self):
        raise NotImplementedError()

    def get_javascript(self):
        raise NotImplementedError

class ComponentDiGraphHTMLRenderer(GraphRenderer):
    def __init__(self,
                 component_graph,
                 width=1000,
                 height=1000,
                 enable_pan_and_zoom=True,
                 engine='dot'):
        self.component_graph = component_graph
        self.width = width
        self.height = height
        self.enable_pan_and_zoom = enable_pan_and_zoom
        self.engine = engine

    def get_graph(self):

        selected_components = self.component_graph.get_components_at_least_size(0)
        all_svg = ''
        for i, component in enumerate(selected_components):
            dot_str = self.component_graph.get_dot(component)
            raw_svg = self.get_svg(dot_str)
            lines = raw_svg.split('\n')
            lines[9] = lines[9].replace('graph0', 'graph%s' % (component))
            lines[6] = ('<svg easypz width="{width}pt" height="{height}pt" id="svg{component}"'
                        + ' style="display: none" class="dotgraph"').format(
                width=self.width, height=self.height, component=component
            )
            for line in lines[6:]:
                if line.startswith('<!--') or line.endswith('-->'):
                    continue
                all_svg += line + '\n'
        return all_svg

    def get_svg(self, dot_str):
        import graphviz as gv

        return gv.Source(dot_str, format='svg', engine=self.engine).pipe().decode('utf-8')

    def get_javascript(self):
        return '''
        name_to_component = %s;
        
        name_to_coord = {}; // term -> {x, y, svg} 
        origViewBox = {}; // svgid -> viewbox
        
        Array.prototype.forEach.call(
            document.querySelectorAll(".dotgraph"),
            function(svg) {
                origViewBox[svg.id] = svg.getAttribute('viewBox');
                
                Array.prototype.forEach.call(
                    svg.getElementsByTagName('text'), 
                    function(text) {
                        name_to_coord[text.textContent] = {
                            "x": text.getAttribute("x"), 
                            "y": text.getAttribute("y"), 
                            "svg": svg.id
                        };
                    }
                )
            }
        ) 
        
        panZoomInstance = null;
        
        
        function zoomToName(name) {
        
            //panZoomInstance.reset();
                        
                        
            console.log("ZOOMING TO "); console.log(name);
            panZoomInstance.fit();
            panZoomInstance.center(); 
            
            var pzSizes = panZoomInstance.getSizes();
            var centerX = name_to_coord[name].x;
            var centerY = -name_to_coord[name].y;
            var newX = centerX*pzSizes["width"]/pzSizes["viewBox"]["width"];
            var newY = centerY*pzSizes["height"]/pzSizes["viewBox"]["height"];
            
            //var zoomRatio = 1/(pzSizes["width"]/pzSizes["viewBox"]["width"]);
            console.log('zr '.concat(name, ' ', pzSizes, ' ', centerX, ' ', centerY, ' ', newX, ' ', newY));       
            panZoomInstance.zoomAtPointBy(5, {'x':newX, 'y':newY});
        }
        
        function panToName(name) {
            var x = name_to_coord[name].x; 
            var y = panZoomInstance.getSizes().viewBox.height + Number.parseInt(name_to_coord[name].y); 
            panZoomInstance.reset() 
            panZoomInstance.zoom(1/panZoomInstance.getSizes().realZoom, true)
            panZoomInstance.pan({x:0,y:0});
            var realZoom = panZoomInstance.getSizes().realZoom; 
            var destX = -((x * realZoom) - (panZoomInstance.getSizes().width/2));
            var destY = -((y * realZoom) - (panZoomInstance.getSizes().height/2));
            panZoomInstance.pan({'x':0,'y':0}); 
            panZoomInstance.pan({'x': destX, 'y': destY})
        }
            
        
        function showTermGraph(term) {
            var nodeName = 'svg' + name_to_component[term];
            document.getElementById(nodeName).style.display='block'; 
            %s
            panToName(term);
        }

        Array.from(document.querySelectorAll('.node')).map(
            function (node) {
                node.addEventListener('mouseenter', mouseEnterNode);
                node.addEventListener('mouseleave', mouseLeaveNode);
                node.addEventListener('click', clickNode);
            }
        )
        
        function clickNode() {
            document.querySelectorAll(".dotgraph")
                .forEach(node => node.style.display = 'none');

            var term = Array.prototype.filter
                .call(this.children, (x => x.tagName === "text"))[0].textContent;

            plotInterface.handleSearchTerm(term, true);
        }

        function mouseEnterNode(event) {
            var term = Array.prototype.filter.call(this.children, (x => x.tagName === "text"))[0].textContent;
            plotInterface.showTooltipSimple(term);
            this.style.fill="red";
        }

        function mouseLeaveNode() {
            plotInterface.tooltip.transition().style('opacity', 0)
            this.style.fill="black";
        }''' % (json.dumps(self.component_graph.get_node_to_component_dict()),
                self._get_pan_and_zoom_js())

    def _get_pan_and_zoom_js(self):
        if self.enable_pan_and_zoom:
            return '''
                panZoomInstance = svgPanZoom('#' + nodeName, {
                    zoomEnabled: true,
                    controlIconsEnabled: true,
                    fit: true,
                    center: true,
                    maxZoom: 100000,
                    minZoom: 0.1
                  });
            '''
        else:
            return ''


import string

import numpy as np
from scattertext.CorpusFromParsedDocuments import CorpusFromParsedDocuments

from scattertext.features.FeatsFromSpacyDoc import FeatsFromSpacyDoc


class CorpusWithoutCategoriesFromParsedDocuments(object):
    def __init__(self,
                 df,
                 parsed_col,
                 feats_from_spacy_doc=FeatsFromSpacyDoc()):

        '''
        Parameters
        ----------
        df : pd.DataFrame
         contains category_col, and parse_col, were parsed col is entirely spacy docs
        parsed_col : str
            name of spacy parsed column in convention_df
        feats_from_spacy_doc : FeatsFromSpacyDoc
        '''
        self.df = df
        self.parsed_col = parsed_col
        self.feats_from_spacy_doc = feats_from_spacy_doc

    def build(self):
        '''

        :return: ParsedCorpus
        '''
        category_col = 'Category'
        while category_col in self.df:
            category_col = 'Category_' + ''.join(np.random.choice(string.ascii_letters) for _ in range(5))
        return CorpusFromParsedDocuments(
            self.df.assign(**{category_col: '_'}),
            category_col,
            self.parsed_col,
            feats_from_spacy_doc=self.feats_from_spacy_doc,
        ).build()


import numpy as np
from scipy.optimize import curve_fit


# from https://stackoverflow.com/questions/55725139/fit-sigmoid-function-s-shape-curve-to-data-using-python
def sigmoid(x, L, x0, k, b):
    y = L / (1 + np.exp(-k * (x - x0))) + b
    return (y)

class Sigmoidal:
    def __init__(self):
        self.popt = None

    def fit(self, x, y):
        assert len(x) == len(y)

        p0 = [max(y), np.median(x), 1, min(y)]  # this is an mandatory initial guess

        self.popt, pcov = curve_fit(sigmoid, x, y, p0, method='dogbox', maxfev=10000)
        return self

    def fit_predict(self, x, y):
        self.fit(x, y)
        return self.predict(x)

    def predict(self, x):
        return sigmoid(x, *self.popt)



'''
All data from
Johan Bollen, Marijn ten Thij, Fritz Breithaupt, Alexander T. J. Barron, Lauren A. Rutter, Lorenzo Lorenzo-Luaces,
Marten Scheffer. Historical language records reveal a surge of cognitive distortions in recent decades.
Proceedings of the National Academy of Sciences Jul 2021, 118 (30) e2102061118; DOI: 10.1073/pnas.210206111
'''
from typing import Dict, List, Optional

from scattertext.features.featoffsets.feat_and_offset_getter import FeatAndOffsetGetter


import flashtext

COGNITIVE_DISTORTIONS_DEFINITIONS = {
    'Catastrophizing': 'Exaggerating the importance of negative events',
    'Dichotomous Reasoning': 'Thinking that an inherently continuous situation can only fall into two categories',
    'Disqualifying the Positive': 'Unreasonably discounting positive experiences',
    'Emotional Reasoning': 'Thinking that something is true based on how one feels, '
                           'ignoring the evidence to the contrary',
    'Fortune-telling': 'Making predictions, usually negative ones, about the future',
    'Labeling and mislabeling': 'Labeling yourself or others while discounting evidence that '
                                'could lead to less disastrous conclusions',
    'Magnification and Minimization': 'Magnifying negative aspects or minimizing positive aspects',
    'Mental Filtering': 'Paying too much attention to negative details instead of the whole picture',
    'Mindreading': 'Believing you know what others are thinking',
    'Overgeneralizing': 'Making sweeping negative conclusions based on a few examples',
    'Personalizing': 'Believing others are behaving negatively because of oneself, '
                     'without considering more plausible or external explanations for behavior',
    'Should statements': 'Having a fixed idea on how you and/or others should behave'}

COGNITIVE_DISTORTIONS_LEXICON = {
    'Catastrophizing': [['will', 'fail'], ['will', 'go', 'wrong'], ['will', 'end'], ['will', 'be', 'impossible'],
                        ['will', 'not', 'happen'], ['will', 'be', 'terrible'], ['will', 'be', 'horrible'],
                        ['will', 'be', 'a', 'catastrophe,'], ['will', 'be', 'a', 'disaster'], ['will', 'never', 'end'],
                        ['will', 'not', 'end']],
    'Dichotomous Reasoning': [['only'], ['every'], ['everyone'], ['everybody'], ['everything'], ['everywhere'],
                              ['always'], ['perfect'], ['the', 'best'], ['all'], ['not', 'a', 'single'], ['no', 'one'],
                              ['nobody,'], ['nothing'], ['nowhere'], ['never'], ['worthless'], ['the', 'worst'],
                              ['neither'], ['nor'], ['either', 'or'], ['black', 'or', 'white'], ['ever']],
    'Disqualifying the Positive': [['great', 'but'], ['good', 'but'], ['OK', 'but'], ['not', 'that', 'great'],
                                   ['not', 'that', 'good'], ['it', 'was', 'not'], ['not', 'all', 'that'],
                                   ['fine', 'but'], ['acceptable', 'but'], ['great', 'yet,'], ['good', 'yet'],
                                   ['OK', 'yet'], ['fine', 'yet'], ['acceptable', 'yet']],
    'Emotional Reasoning': [['but', 'I', 'feel'], ['since', 'I', 'feel'], ['because', 'I', 'feel'],
                            ['but', 'it', 'feels'], ['since', 'it', 'feels'], ['because', 'it', 'feels'],
                            ['still', 'feels']],
    'Fortune-telling': [['I', 'will', 'not'], ['we', 'will', 'not'], ['you', 'will', 'not'], ['they', 'will', 'not'],
                        ['it', 'will', 'not'], ['that', 'will', 'not'], ['he', 'will', 'not'], ['she', 'will', 'not']],
    'Labeling and mislabeling': [['I', 'am', 'a'], ['he', 'is', 'a'], ['she', 'is', 'a'], ['they', 'are', 'a'],
                                 ['it', 'is', 'a'], ['that', 'is', 'a'], ['sucks', 'at'], ['suck', 'at'],
                                 ['I', 'never'], ['he', 'never'], ['she', 'never'], ['you', 'never'], ['we', 'never,'],
                                 ['they', 'never'], ['I', 'am', 'an'], ['he', 'is', 'an'], ['she', 'is', 'an'],
                                 ['they', 'are', 'an'], ['it', 'is', 'an'], ['that', 'is', 'an'], ['a', 'burden'],
                                 ['a', 'complete'], ['a', 'completely'], ['a', 'huge,'], ['a', 'loser'], ['a', 'major'],
                                 ['a', 'total'], ['a', 'totally'], ['a', 'weak'], ['an', 'absolute'], ['an', 'utter'],
                                 ['a', 'bad'], ['a', 'broken'], ['a', 'damaged'], ['a', 'helpless'], ['a', 'hopeless'],
                                 ['an', 'incompetent'], ['a', 'toxic'], ['an', 'ugly'], ['an', 'undesirable'],
                                 ['an', 'unlovable'], ['a', 'worthless'], ['a', 'horrible'], ['a', 'terrible']],
    'Magnification and Minimization': [['worst'], ['best'], ['not', 'important'], ['not', 'count'], ['not', 'matter'],
                                       ['no', 'matter'], ['the', 'only', 'thing'], ['the', 'one', 'thing']],
    'Mental Filtering': [['I', 'see', 'only'], ['all', 'I', 'see'], ['all', 'I', 'can', 'see'],
                         ['can', 'only', 'think'], ['nothing', 'good'], ['nothing', 'right'], ['completely', 'bad'],
                         ['completely', 'wrong'], ['only', 'the', 'bad'], ['only', 'the', 'worst'], ['if', 'I', 'just'],
                         ['if', 'I', 'only'], ['if', 'it', 'just'], ['if', 'it', 'only']],
    'Mindreading': [['everyone', 'believes'], ['everyone', 'knows'], ['everyone', 'thinks'],
                    ['everyone', 'will', 'believe'], ['everyone', 'will', 'know'], ['everyone', 'will', 'think,'],
                    ['nobody', 'believes'], ['nobody', 'knows'], ['nobody', 'thinks'], ['nobody', 'will', 'believe'],
                    ['nobody', 'will', 'know'], ['nobody', 'will', 'think'], ['he', 'believes,'], ['he', 'knows'],
                    ['he', 'thinks'], ['he', 'does', 'not', 'believe'], ['he', 'does', 'not', 'know'],
                    ['he', 'does', 'not', 'think'], ['he', 'will', 'believe'], ['he', 'will', 'know'],
                    ['he', 'will', 'think'], ['he', 'will', 'not', 'believe'], ['he', 'will', 'not', 'know'],
                    ['he', 'will', 'not', 'think'], ['she', 'believes'], ['she', 'knows'], ['she', 'thinks'],
                    ['she', 'does', 'not', 'believe,'], ['she', 'does', 'not', 'know'], ['she', 'does', 'not', 'think'],
                    ['she', 'will', 'believe'], ['she', 'will', 'know'], ['she', 'will', 'think'],
                    ['she', 'will', 'not', 'believe'], ['she', 'will', 'not', 'know'], ['she', 'will', 'not', 'think'],
                    ['they', 'believe'], ['they', 'know'], ['they', 'think'], ['they', 'do', 'not', 'believe'],
                    ['they', 'do', 'not', 'know'], ['they', 'do', 'not', 'think,'], ['they', 'will', 'believe'],
                    ['they', 'will', 'know'], ['they', 'will', 'think'], ['they', 'will', 'not', 'believe'],
                    ['they', 'will', 'not', 'know'], ['they', 'will', 'not', 'think'], ['we', 'believe'],
                    ['we', 'know'], ['we', 'think'], ['we', 'do', 'not', 'believe'], ['we', 'do', 'not', 'know'],
                    ['we', 'do', 'not', 'think'], ['we', 'will', 'believe'], ['we', 'will', 'know'],
                    ['we', 'will', 'think'], ['we', 'will', 'not', 'believe'], ['we', 'will', 'not', 'know'],
                    ['we', 'will', 'not', 'think'], ['you', 'believe'], ['you', 'know'], ['you', 'think'],
                    ['you', 'do', 'not', 'believe'], ['you', 'do', 'not', 'know,'], ['you', 'do', 'not', 'think'],
                    ['you', 'will', 'believe'], ['you', 'will', 'know'], ['you', 'will', 'think'],
                    ['you', 'will', 'not', 'believe'], ['you', 'will', 'not', 'know'], ['you', 'will', 'not', 'think']],
    'Overgeneralizing': [['all', 'of', 'the', 'time'], ['all', 'of', 'them'], ['all', 'the', 'time'],
                         ['always', 'happens'], ['always', 'like'], ['happens', 'every', 'time'], ['completely'],
                         ['no', 'one', 'ever,'], ['nobody', 'ever'], ['every', 'single', 'one', 'of', 'them'],
                         ['every', 'single', 'one', 'of', 'you'], ['I', 'always'], ['you', 'always'], ['he', 'always'],
                         ['she', 'always'], ['they', 'always'], ['I', 'am', 'always'], ['you', 'are', 'always'],
                         ['he', 'is', 'always'], ['she', 'is', 'always'], ['they', 'are', 'always']],
    'Personalizing': [['all', 'me'], ['all', 'my'], ['because', 'I'], ['because', 'my'], ['because', 'of', 'my'],
                      ['because', 'of', 'me'], ['I', 'am', 'responsible'], ['blame', 'me'], ['I', 'caused,'],
                      ['I', 'feel', 'responsible'], ['all', 'my', 'doing'], ['all', 'my', 'fault'], ['my', 'bad'],
                      ['my', 'responsibility']],
    'Should statements': [['should'], ['ought'], ['must'], ['have', 'to'], ['has', 'to']]}


class LexiconFeatAndOffsetGetter(FeatAndOffsetGetter):
    def __init__(
            self,
            lexicons: Dict[str, List[List[str]]],
            definitions: Optional[Dict[str, str]] = None,
            case_sensitive: bool = False
    ):
        '''
        :param lexicons: Dict[str, List[List[str]]], maps a topic name to its list of word sequences
        :param definitions, Optional[Dict[str, str]], maps topic name to an explanation.
        :param case_sensitive: bool, are word sequences case sensitive?
        '''
        self.definitions = definitions
        self.keyword_processor = flashtext.KeywordProcessor(case_sensitive)
        self.keyword_processor.add_keywords_from_dict({
            topic: [' '.join(term_sequence) for term_sequence in term_sequences]
            for topic, term_sequences in lexicons.items()
        })

    def get_term_offsets(self, doc):
        return []

    def get_metadata_offsets(self, doc):
        text = str(doc)
        offset_tokens = {}
        for topic, start_index, end_index in self.keyword_processor.extract_keywords(text, span_info=True):
            token_stats = offset_tokens.setdefault(topic, [0, []])
            token_stats[0] += 1
            token_stats[1].append((start_index, end_index))
        return list(offset_tokens.items())

    def get_definitions(self):
        return self.definitions





import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph._traversal import connected_components
from scattertext.graphs.ComponentDiGraph import ComponentDiGraph


class SimpleDiGraph(object):
    def __init__(self, orig_edge_df):
        self.orig_edge_df = orig_edge_df
        self.id_node_df = pd.DataFrame({'name': list(set(orig_edge_df.source.values)
                                                     | set(orig_edge_df.target.values))})
        self.node_df = self.id_node_df.reset_index().set_index('name')
        self.edge_df = pd.merge(
            pd.merge(
                self.orig_edge_df, self.node_df, left_on='source', right_index=True
            ).rename(columns={'index': 'source_id'}),
            self.node_df,
            left_on='target',
            right_index=True
        ).rename(columns={'index': 'target_id'})

    def make_component_digraph(self, graph_params, node_params):
        components = self.get_connected_components()
        return ComponentDiGraph(
            edge_df=self.edge_df.assign(Component=lambda df: components[df.source_id]),
            orig_edge_df=self.orig_edge_df.assign(Component=components[self.edge_df.source_id]),
            node_df=self.node_df.assign(Component=lambda df: components[df['index']]),
            id_node_df=self.id_node_df.assign(Component=lambda df: components[df.index]),
            components=components,
            graph_params=graph_params,
            node_params=node_params
        )

    def get_connected_subgraph_df(self):
        return pd.DataFrame({'component': self.get_connected_components(),
                             'nodes': self.node_df['index'].values,
                             'values': self.node_df.index}
                            ).groupby('component').agg(list).assign(
            size=lambda df: df.nodes.apply(len)
        ).sort_values(by='size', ascending=False).reset_index(drop=True)

    def get_connected_components(self):
        return connected_components(self.get_sparse_graph(), True)[1]

    def get_sparse_graph(self):
        return csr_matrix(
            (np.ones(len(self.edge_df)),
             (self.edge_df.source_id.values, self.edge_df.target_id.values)),
            shape=(len(self.node_df), len(self.node_df))
        )

    def limit_nodes(self, nodes):
        return SimpleDiGraph(self.edge_df[
                                 self.edge_df.target_id.isin(nodes) | self.edge_df.source_id.isin(nodes)
                                 ][['source', 'target']])



from unittest import TestCase

import numpy as np
import pandas as pd

from scattertext import LogOddsRatioInformativeDirichletPrior
from scattertext.PriorFactory import PriorFactory
from scattertext.test.test_semioticSquare import get_test_corpus


class TestPriorFactory(TestCase):
    def test_all_categories(self):
        corpus = get_test_corpus()
        priors, my_corpus = (PriorFactory(corpus, starting_count=0, category='hamlet')
                             .use_all_categories()
                             .build())
        tdf = corpus.get_term_freq_df()
        self.assertEqual(len(priors), len(tdf))
        np.testing.assert_equal(priors.values,
                                corpus.get_term_freq_df().sum(axis=1).values)

    def test_neutral_categories(self):
        corpus = get_test_corpus()
        priors = (PriorFactory(corpus, 'hamlet', starting_count=0.001,
                               not_categories=['swift'])
                  .use_neutral_categories()
                  .get_priors())
        self.assertEqual(priors.min(), 0.001)
        self.assertEqual(priors.shape[0], corpus._X.shape[1])

        corpus = get_test_corpus()
        priors = (PriorFactory(corpus, 'hamlet', starting_count=0.001,
                               not_categories=['swift'])
                  .use_neutral_categories()
                  .drop_zero_priors()
                  .get_priors())

        jzcnts = corpus.get_term_freq_df()['jay-z/r. kelly freq'].where(lambda x: x > 0).dropna()
        np.testing.assert_equal(priors.values,
                                jzcnts.values + 0.001)

    def test_get_general_term_frequencies(self):
        corpus = get_test_corpus()
        fact = (PriorFactory(corpus,
                             category='hamlet',
                             not_categories=['swift'],
                             starting_count=0)
                .use_general_term_frequencies()
                .use_all_categories()
                )
        priors, clean_corpus = fact.build()

        expected_prior = pd.merge(corpus.get_term_doc_count_df(),
                                  corpus.get_term_and_background_counts()[['background']],
                                  left_index=True,
                                  right_index=True,
                                  how='left').fillna(0.).sum(axis=1)

        np.testing.assert_allclose(priors.values, expected_prior.values)

    def test_align_to_target(self):
        full_corpus = get_test_corpus()
        corpus = full_corpus.remove_categories(['swift'])
        priors = PriorFactory(full_corpus).use_all_categories().get_priors()
        with self.assertRaises(ValueError):
            (LogOddsRatioInformativeDirichletPrior(priors)
             .get_scores(*corpus.get_term_freq_df().values.T))
        priors = (PriorFactory(full_corpus)
                  .use_all_categories()
                  .align_to_target(corpus)
                  .get_priors())
        (LogOddsRatioInformativeDirichletPrior(priors)
         .get_scores(*corpus.get_term_freq_df().values.T))

    def test_use_categories(self):
        full_corpus = get_test_corpus()
        priors = PriorFactory(full_corpus).use_categories(['swift']).get_priors()
        corpus = full_corpus.remove_categories(['swift'])
        with self.assertRaises(ValueError):
            (LogOddsRatioInformativeDirichletPrior(priors)
             .get_scores(*corpus.get_term_freq_df().values.T))
        priors = (PriorFactory(full_corpus)
                  .use_all_categories()
                  .align_to_target(corpus)
                  .get_priors())
        (LogOddsRatioInformativeDirichletPrior(priors)
         .get_scores(*corpus.get_term_freq_df().values.T))

    def test_get_custom_term_frequencies(self):
        corpus = get_test_corpus()
        fact = (PriorFactory(corpus, starting_count=0.04)
                .use_custom_term_frequencies(pd.Series({'halt': 3, 'i': 8}))
                .drop_zero_priors()
                )
        priors, clean_corpus = fact.build()
        self.assertEqual(set(clean_corpus.get_terms()), {'i', 'halt'})
        np.testing.assert_equal(priors.sort_values().values, [3.04, 8.04])



from scattertext.features.FeatsFromSpacyDocAndEmpath import FeatsFromSpacyDocAndEmpath
from collections import Counter


class FeatsFromOnlyEmpath(FeatsFromSpacyDocAndEmpath):
	def get_feats(self, doc):
		return Counter()
	def get_doc_metadata(self, doc, prefix=''):
		return FeatsFromSpacyDocAndEmpath.get_doc_metadata(self, doc, prefix=prefix)





from scattertext.indexstore import IndexStore

from scattertext.Corpus import Corpus


class DataFrameCorpus(Corpus):
    def __init__(self,
                 X,
                 mX,
                 y,
                 term_idx_store,
                 category_idx_store,
                 metadata_idx_store,
                 raw_texts,
                 df,
                 unigram_frequency_path=None):
        '''
        Parameters
        ----------
        X : csr_matrix
            term document matrix
        mX : csr_matrix
            metadata-document matrix
        y : np.array
            category index array
        term_idx_store : IndexStore
            Term indices
        category_idx_store : IndexStore
            Catgory indices
        metadata_idx_store : IndexStore
          Document metadata indices
        raw_texts : np.array or pd.Series
            Raw texts
        unigram_frequency_path : str or None
            Path to term frequency file.
        '''
        self._df = df
        Corpus.__init__(self, X, mX, y,
                        term_idx_store,
                        category_idx_store,
                        metadata_idx_store,
                        raw_texts,
                        unigram_frequency_path)

    def _apply_mask_to_df(self, new_y_mask, new_df):
        df_to_ret = self._df if new_df is None else new_df
        if new_y_mask is None:
            return df_to_ret

        return df_to_ret[new_y_mask].reset_index(drop=True)

    def get_df(self):
        '''
        Returns
        -------
        pd.DataFrame
        '''
        return self._df

    def get_field(self, field):
        '''
        Parameters
        ----------
        field: str, field name

        Returns
        -------
        pd.Series, all members of field
        '''
        return self._df[field]

    def assign(self, **kwargs):
        '''
        Runs assign in the internal dataframe

        :param kwargs:
        :return: Corpus
        '''
        self._df = self._df.assign(**kwargs)
        return self

    def search(self, ngram):
        '''
        Parameters
        ----------
        ngram, str or unicode, string to search for

        Returns
        -------
        pd.DataFrame, {self._parsed_col: <matching texts>, self._category_col: <corresponding categories>, ...}

        '''
        mask = self._document_index_mask(ngram)
        return self._df[mask]

    def make_column_metadata(self, column):
        '''

        :param column: str
        :return: Corpus
        '''

        return self.add_doc_names_as_metadata(self.get_df()[column])



from typing import Optional, List
import pandas as pd
import numpy as np
from scipy.sparse import csr_matrix

from scattertext.ParsedCorpus import ParsedCorpus
from scattertext.TermDocMatrix import TermDocMatrix
from scattertext.CorpusDF import CorpusDF
from scattertext.indexstore import IndexStore, IndexStoreFromList


class CorpusFromTermFrequencies(object):
    def __init__(self,
                 X: csr_matrix,
                 term_vocabulary: List[str],
                 mX: Optional[csr_matrix] = None,
                 y: Optional[np.array] = None,
                 category_names: Optional[str] = None,
                 metadata_vocabulary: Optional[List[str]] = None,
                 text_df: Optional[pd.DataFrame] = None,
                 text_col: Optional[str] = None,
                 parsed_col: Optional[str] = None,
                 category_col: Optional[str] = None,
                 unigram_frequency_path: Optional[str] = None):
        '''
        Parameters
        ----------
        X: csr_matrix; term-document frequency matrix; columns represent terms and rows documents
        term_vocabulary: List[str]; Each entry corresponds to a term
        mX: Optional[csr_matrix]; metadata csr matrix
        y: Optional[np.array[int]]; indices of category names for each document
        category_names: Optional[List[str]], names of categories for y
        text_df: pd.DataFrame with a row containing the raw document text
        text_col: str; name of row containing the text of each document
        parsed_col: str; name of row containing the parsed text of each document
        unigram_frequency_path: str (see TermDocMatrix)

        '''
        self.X = X
        self.term_idx_store = IndexStoreFromList.build(term_vocabulary)
        assert self.X.shape[1] == len(term_vocabulary)
        self.metadata_idx_store = IndexStore()
        if y is None:
            self.y = np.zeros(self.X.shape[0], dtype=int)
            self.category_idx_store = IndexStoreFromList.build(['_'])
            assert category_names is None
        else:
            self.y = y
            assert len(category_names) == len(set(y))
            self.category_idx_store = IndexStoreFromList.build(category_names)
        if metadata_vocabulary is not None:
            assert mX.shape[1] == metadata_vocabulary
            self.mX = mX
            self.metadata_idx_store = IndexStoreFromList.build(metadata_vocabulary)
        else:
            assert metadata_vocabulary is None
            self.mX = csr_matrix((0, 0))
            self.metadata_idx_store = IndexStore()
        self.text_df = text_df
        if parsed_col is not None:
            assert parsed_col in text_df
        if text_col is not None:
            assert text_col in text_df
        if category_col is not None:
            assert category_col in text_df
        self.category_col = category_col
        self.text_col = text_col
        self.parsed_col = parsed_col
        self.unigram_frequency_path = unigram_frequency_path

    def build(self):
        '''
        Returns
        -------
        CorpusDF
        '''
        if self.text_df is not None:
            if self.parsed_col is not None:
                if self.category_col is None:
                    self.text_df = self.text_df.assign(Category=self.category_idx_store.getvalbatch(self.y))
                    self.category_col = 'Category'
                return ParsedCorpus(
                    df=self.text_df,
                    X=self.X,
                    mX=self.mX,
                    y=self.y,
                    parsed_col=self.parsed_col,
                    term_idx_store=self.term_idx_store,
                    category_idx_store=self.category_idx_store,
                    metadata_idx_store=self.metadata_idx_store,
                    unigram_frequency_path=self.unigram_frequency_path,
                    category_col=self.category_col
                )
            elif self.text_col is not None:
                return CorpusDF(
                    df=self.text_df,
                    X=self.X,
                    mX=self.mX,
                    y=self.y,
                    text_col=self.text_col,
                    term_idx_store=self.term_idx_store,
                    category_idx_store=self.category_idx_store,
                    metadata_idx_store=self.metadata_idx_store,
                    unigram_frequency_path=self.unigram_frequency_path
                )
        return TermDocMatrix(
            X=self.X,
            mX=self.mX,
            y=self.y,
            term_idx_store=self.term_idx_store,
            category_idx_store=self.category_idx_store,
            metadata_idx_store=self.metadata_idx_store,
            unigram_frequency_path=self.unigram_frequency_path
        )



from collections import Counter
from unittest import TestCase

from scattertext.features.FeatsFromTopicModel import FeatsFromTopicModel


class TestFeatsFromTopicModel(TestCase):
    def test_get_doc_get_feats(self):
        expected = Counter({'a b': 2, 'c e f': 1, 'b': 1})

        actual = FeatsFromTopicModel(
            topic_model={'Topic A': ['A b', 'b', 'C e F'],
                         'Topic B': ['B', 'C e F']},
        ).get_feats('A b A b C e F B')

        self.assertEqual(expected, actual)


    def test_get_doc_metadata(self):
        expected = Counter({'Topic A': 3, 'Topic B': 2})

        actual = FeatsFromTopicModel(
            topic_model={'Topic A': ['A b', 'b', 'C e F'],
                         'Topic B': ['B', 'C e F']},
            keyword_processor_args={'case_sensitive': True}
        ).get_doc_metadata('A b A b C e F B')

        self.assertEqual(expected, actual)



import html

import pandas as pd

def get_tooltip_js_function(plot_df, tooltip_column_names, tooltip_columns):
    if len(tooltip_columns) > 2:
        raise Exception("You can have at most two columns in a tooltip.")
    tooltip_content = ''
    tooltip_column_names = {} if tooltip_column_names is None else tooltip_column_names
    for col in tooltip_columns:
        assert col in plot_df
        formatting = ''
        if pd.api.types.is_float(plot_df[col].iloc[0]):
            formatting = '.toFixed(6)'
        tooltip_content += '+ "<br />%s: " + d.etc["%s"]%s' % (
            html.escape(tooltip_column_names.get(col, col)),
            col.replace('"', '\\"').replace("'", "\\'"), formatting)
    tooltip_content_js_function = '(function(d) {return d.term %s;})' % tooltip_content
    return tooltip_content_js_function


def get_custom_term_info_js_function(plot_df, term_description_column_names, term_description_columns,
                                     term_word_in_term_description):
    custom_term_html = ''
    term_description_column_names = {} if term_description_column_names is None else term_description_column_names
    for col in term_description_columns:
        assert col in plot_df
        formatting = '.toFixed(6)' if pd.api.types.is_float(plot_df[col].iloc[0]) else ''
        custom_term_html += '+ "<b>%s:</b> " + d.etc["%s"]%s + "<br />"' % (
            html.escape(term_description_column_names.get(col, col)),
            col.replace('"', '\\"').replace("'", "\\'"), formatting)
    if custom_term_html != '':
        custom_term_html += '+'
    custom_term_info_js_function = (
            '(d => "%s: "+d.term+"<div class=topic_preview>"%s"</div>")' %
            (
                term_word_in_term_description,
                custom_term_html
            )
    )
    return custom_term_info_js_function


from scattertext.features.featoffsets.feat_and_offset_getter import FeatAndOffsetGetter


class TokenFeatAndOffsetGetter(FeatAndOffsetGetter):
    def get_term_offsets(self, doc):
        return []

    def get_metadata_offsets(self, doc):
        offset_tokens = {}
        for sent in doc.sents:
            for tok in sent:
                token_stats = offset_tokens.setdefault(tok.lower_, [0, []])
                token_stats[0] += 1
                token_stats[1].append((tok.idx, tok.idx + len(tok.lower_)))
        return offset_tokens.items()



import numpy as np
import pandas as pd
from scipy.stats import rankdata


def scale(vec, other_vec=None):
    if other_vec is None:
        other_vec = vec
    return (other_vec - vec.min()) / (vec.max() - vec.min())


class Coordinates:
    def __init__(self, x, y):
        self.x = x
        self.y = y

def scale_jointly(x, y):
    x, y = np.array(x), np.array(y)
    global_max, global_min = max(x.max(), y.max()), min(x.min(), y.min())
    x_scaled = np.zeros(len(x), dtype=float)
    y_scaled = np.zeros(len(y), dtype=float)
    x_scaled[x < 0] = -(x[x < 0] / global_min)
    x_scaled[x > 0] = (x[x > 0] / global_max)
    y_scaled[y < 0] = -(y[y < 0] / global_min)
    y_scaled[y > 0] = (y[y > 0] / global_max)
    return Coordinates((x_scaled + 1) / 2, (y_scaled + 1) / 2)


def rotate_degrees(x, y, degrees):
    return rotate_radians(y, x, np.pi * (degrees) / 180)


def rotate_radians(y, x, radians):
    y = np.array(y)
    x = np.array(x)
    return Coordinates(
        x * np.cos(radians) - y * np.sin(radians),
        x * np.sin(radians) + y * np.cos(radians)
    )

def scale_center_zero(vec, other_vec=None):
    if other_vec is None:
        other_vec = vec
    return ((((other_vec > 0).astype(float) * (other_vec / vec.max())) * 0.5 + 0.5)
            + ((other_vec < 0).astype(float) * (other_vec / (-vec.min())) * 0.5))

def scale_center_zero_abs(vec, other_vec=None):
    if other_vec is None:
        other_vec = vec
    max_abs = max(vec.max(), -vec.min())
    return ((((other_vec > 0).astype(float) * (other_vec / max_abs)) * 0.5 + 0.5)
            + ((other_vec < 0).astype(float) * (other_vec / max_abs) * 0.5))


def scale_neg_1_to_1_with_zero_mean_abs_max(vec):
    max_abs = max(vec.max(), -vec.min())
    return ((((vec > 0).astype(float) * (vec / max_abs)) * 0.5 + 0.5)
            + ((vec < 0).astype(float) * (vec / max_abs) * 0.5))


def scale_neg_1_to_1_with_zero_mean(vec):
    return ((((vec > 0).astype(float) * (vec / vec.max())) * 0.5 + 0.5)
            + ((vec < 0).astype(float) * (vec / (-vec.min())) * 0.5))


def scale_neg_1_to_1_with_zero_mean_rank_abs_max(v):
    rankv = v * 2 - 1
    pos_v = rankv[rankv > 0]
    pos_v = rankdata(pos_v, 'dense')
    pos_v = pos_v / pos_v.max()
    neg_v = rankv[rankv < 0]
    neg_v = rankdata(neg_v, 'dense')
    neg_v = neg_v / neg_v.max()
    rankv[rankv > 0] = pos_v
    rankv[rankv < 0] = - (neg_v.max() - neg_v)

    return scale_neg_1_to_1_with_zero_mean_abs_max(rankv)


def scale_neg_1_to_1_with_zero_mean_log_abs_max(v):
    '''
    !!! not working
    '''
    df = pd.DataFrame({'v': v,
                       'sign': (v > 0) * 2 - 1})
    df['lg'] = np.log(np.abs(v)) / np.log(1.96)
    df['exclude'] = (np.isinf(df.lg) | np.isneginf(df.lg))
    for mask in [(df['sign'] == -1) & (df['exclude'] == False),
                 (df['sign'] == 1) & (df['exclude'] == False)]:
        df[mask]['lg'] = df[mask]['lg'].max() - df[mask]['lg']
    df['lg'] *= df['sign']
    df['lg'] = df['lg'].fillna(0)
    print(df[df['exclude']]['lg'].values)
    # to_rescale = convention_df['lg'].reindex(v.index)
    df['to_out'] = scale_neg_1_to_1_with_zero_mean_abs_max(df['lg'])
    print('right')
    print(df.sort_values(by='lg').iloc[:5])
    print(df.sort_values(by='lg').iloc[-5:])
    print('to_out')
    print(df.sort_values(by='to_out').iloc[:5])
    print(df.sort_values(by='to_out').iloc[-5:])
    print(len(df), len(df.dropna()))
    return df['to_out']


def scale_standardize(vec, terms=None, other_vec=None):
    to_ret = (vec - vec.mean()) / vec.std()
    to_ret += to_ret.min() + 1
    # to_ret = np.log(to_ret)/np.log(2)
    to_ret += to_ret.min()
    return to_ret


def log_scale_standardize(vec, terms=None, other_vec=None):
    vec_ss = (np.log(vec + 1) / np.log(2))
    vec_ss = (vec_ss - vec_ss.mean()) / vec_ss.std()
    return scale_0_to_1(vec_ss)


def sqrt_scale_standardize(vec, terms=None, other_vec=None):
    vec_ss = np.sqrt(vec)
    vec_ss = (vec_ss - vec_ss.mean()) / vec_ss.std()
    return scale_0_to_1(vec_ss)


def power_scale_standardize_factory(alpha):
    def f(vec, terms=None, other_vec=None):
        vec_ss = np.power(vec, 1. / alpha)
        vec_ss = (vec_ss - vec_ss.mean()) / vec_ss.std()
        return scale_0_to_1(vec_ss)

    return f


'''
from statsmodels.distributions import ECDF
def ecdf_scale_standardize(vec):
	vec_ss = ECDF(vec)(vec)
	vec_ss = (vec_ss - vec_ss.min()) / (vec_ss.max() - vec_ss.min())
	return vec_ss
'''


def power_scale_factory(alpha):
    def f(vec, terms=None, other_vec=None):
        return scale_0_to_1(np.power(vec, 1. / alpha))

    return f


def sqrt_scale(vec, terms=None, other_vec=None):
    return scale_0_to_1(np.sqrt(vec))


def log_scale(vec, terms=None, other_vec=None):
    return scale_0_to_1(np.log(vec))


def percentile(vec, terms=None, other_vec=None):
    vec_ss = rankdata(vec, method='average') * (1. / len(vec))
    return scale_0_to_1(vec_ss)


def dense_rank(vec, terms=None, other_vec=None):
    vec_ss = rankdata(vec, method='dense') * (1. / len(vec))
    return scale_0_to_1(vec_ss)


percentile_dense = dense_rank


def percentile_ordinal(vec, terms=None, other_vec=None):
    vec_ss = rankdata(vec, method='ordinal') * (1. / len(vec))
    return scale_0_to_1(vec_ss)


def percentile_min(vec, terms=None, other_vec=None):
    vec_ss = rankdata(vec, method='min') * (1. / len(vec))
    return scale_0_to_1(vec_ss)


def percentile_alphabetical(vec, terms, other_vec=None):
    scale_df = pd.DataFrame({'scores': vec, 'terms': terms})
    if other_vec is not None:
        scale_df['others'] = other_vec
    else:
        scale_df['others'] = 0
    vec_ss = (scale_df
              .sort_values(by=['scores', 'terms', 'others'],
                           ascending=[True, True, False])
              .reset_index()
              .sort_values(by='index')
              .index)
    return scale_0_to_1(vec_ss)


def stretch_0_to_1(vec):
    a = stretch_neg1_to_1(vec)
    return 0.5 * (a + 1)


def stretch_neg1_to_1(vec):
    a = np.copy(vec)
    if sum(a < 0):
        a[a < 0] = -a[a < 0] * 1. / a[a < 0].min()
    if sum(a > 0):
        a[a > 0] = a[a > 0] * 1. / a[a > 0].max()
    return a


def scale_0_to_1(vec_ss):
    vec_ss = (vec_ss - vec_ss.min()) * 1. / (vec_ss.max() - vec_ss.min())
    return vec_ss



import re

from scattertext.emojis.ProcessedEmojiStructure import VALID_EMOJIS

'''
This is a fast but awful partial implementation of Spacy.   It's useful 
for testing and when you don't need POS tagging. 
'''


class Tok:
    def __init__(self, pos, lem, token, ent_type, tag, is_punct=False, idx=None):
        self.pos_ = pos
        self.lemma_ = lem
        self.lower_ = token.lower()
        self.orth_ = token
        self.ent_type_ = ent_type
        self.tag_ = tag
        self.is_punct = is_punct
        self.idx = idx

    def __str__(self):
        return self.lower_

    def __len__(self):
        return len(self.lower_)


class Doc:
    def __init__(self, sents, raw=None, noun_chunks=[]):
        self.sents = sents
        if raw is None:
            self.string = ' '.join(
                ''.join([tok.lower_ for tok in sent]) for sent in sents
            )
        else:
            self.string = raw
        self.text = self.string
        self.noun_chunks = noun_chunks
        self.toks = []
        for sent in sents:
            self.toks += sent

    def __str__(self):
        return self.string

    def __repr__(self):
        return self.string

    def __iter__(self):
        for sent in self.sents:
            for tok in sent:
                yield tok

    def __getitem__(self, idx):
        return self.toks[idx]  # to do: make this more efficient by only having self.sents

    def __len__(self):
        return len(self.toks)


def whitespace_nlp(doc, entity_type=None, tag_type=None):
    toks = _regex_parse_sentence(doc, entity_type, tag_type)
    return Doc([toks])


def _regex_parse_sentence(doc, entity_type, tag_type):
    toks = _toks_from_sentence(doc, entity_type, tag_type)
    return toks


WHITESPACE_SPLITTER = re.compile(r"(\W)")


def _toks_from_sentence(doc, entity_type, tag_type):
    toks = []
    for tok in WHITESPACE_SPLITTER.split(doc):
        if len(tok) > 0:
            toks.append(Tok(_get_pos_tag(tok),
                            tok[:2].lower(),
                            tok.lower(),
                            ent_type='' if entity_type is None else entity_type.get(tok, ''),
                            tag='' if tag_type is None else tag_type.get(tok, '')))
    return toks


def nltk_tokenzier_factory(nltk_tokenizer):
    '''
    Parameters
    ----------
    nltk_tokenizer : nltk.tokenize.* instance (e.g., nltk.TreebankWordTokenizer())

    Returns
    -------
    Doc of tweets

    Notes
    -------
    Requires NLTK to be installed
    '''

    def tokenize(text):
        toks = []
        for tok in nltk_tokenizer.tokenize(text):
            if len(tok) > 0:
                toks.append(Tok(_get_pos_tag(tok),
                                tok.lower(),
                                tok.lower(),
                                ent_type='',
                                tag=''))
        return Doc([toks], text)

    return tokenize


def tweet_tokenizier_factory(tweet_tokenizer):
    '''
    Parameters
    ----------
    tweet_tokenizer

    Doc of tweets

    Notes
    -------
    Requires NLTK to be installed :(

    '''
    return nltk_tokenzier_factory(tweet_tokenizer)


PUNCT_MATCHER = re.compile('^\W+$')


def _get_pos_tag(tok):
    pos = 'WORD'
    if tok.strip() == '':
        pos = 'SPACE'
    elif ord(tok[0]) in VALID_EMOJIS:
        pos = 'EMJOI'
    elif PUNCT_MATCHER.match(tok):
        pos = 'PUNCT'
    return pos


DEFAULT_TOK_SPLITTER_RE = re.compile(r'(\W)')


def whitespace_nlp_with_sentences(doc,
                                  entity_type=None,
                                  tag_type=None,
                                  tok_splitter_re=DEFAULT_TOK_SPLITTER_RE):
    sentence_split_pat = re.compile(r'([^\.!?]*?[\.!?$])', re.M)
    sents = []
    raw_sents = sentence_split_pat.findall(doc)
    if len(raw_sents) == 0:
        raw_sents = [doc]
    sent_start_idx = 0
    for sentence in raw_sents:
        toks = []
        start_idx_in_sentence = 0
        for tok in tok_splitter_re.split(sentence):
            if len(tok.strip()) > 0:
                toks.append(Tok(_get_pos_tag(tok),
                                tok[:2].lower(),
                                tok.lower(),
                                ent_type='' if entity_type is None else entity_type.get(tok, ''),
                                tag='' if tag_type is None else tag_type.get(tok, ''),
                                idx=sent_start_idx + start_idx_in_sentence))
            start_idx_in_sentence += len(tok)
        sents.append(toks)
        sent_start_idx += len(sentence)
    return Doc(sents, doc)





import pandas as pd
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.linear_model import RidgeCV, ElasticNetCV, LassoCV
from sklearn.pipeline import Pipeline

from scattertext.continuous.coefficientbase import CoefficientBase


class RidgeCoefficients(CoefficientBase):
    def get_coefficient_df(self, corpus, document_scores, pipeline=None):
        '''

        :param corpus: TermDocMatrix, should just have unigrams
        :param document_scores: np.array, continuous value for each document score
        :return: pd.DataFrame
        '''
        assert document_scores.shape == (corpus.get_num_docs(),)
        tdm = self._get_tdm(corpus)

        model = self._get_default_pipeline(pipeline)
        model.fit(X=tdm.toarray(), y=document_scores)

        df = pd.DataFrame({
            'Word': self._get_terms(corpus),
            'Beta': model.named_steps['lr'].coef_,
            'Frequency': self._get_tdm(corpus).sum(axis=0)[0].A1
        }).set_index('Word')

        return df

    def _get_default_pipeline(self, pipeline):
        return Pipeline([
            ('tfidf', TfidfTransformer(sublinear_tf=True)),
            ('lr', (RidgeCV()))
        ]) if pipeline is None else pipeline



from copy import copy

from scattertext import CSRMatrixFactory
from scattertext.DataFrameCorpus import DataFrameCorpus
from scattertext.ParsedCorpus import ParsedDataFrameCorpus


class OffsetCorpus(ParsedDataFrameCorpus):
    def __init__(self,
                 df,
                 X,
                 mX,
                 y,
                 term_idx_store,
                 category_idx_store,
                 metadata_idx_store,
                 parsed_col,
                 category_col,
                 term_offsets,
                 metadata_offsets,
                 unigram_frequency_path=None):
        self._term_offsets = term_offsets
        self._metadata_offsets = metadata_offsets
        ParsedDataFrameCorpus.__init__(self, parsed_col, category_col)
        DataFrameCorpus.__init__(self, X, mX, y, term_idx_store, category_idx_store,
                                 metadata_idx_store,
                                 df[self._parsed_col],
                                 df,
                                 unigram_frequency_path)

    def get_offsets(self):
        return self._metadata_offsets

    def _make_new_term_doc_matrix(self,
                                  new_X=None,
                                  new_mX=None,
                                  new_y=None,
                                  new_term_idx_store=None,
                                  new_category_idx_store=None,
                                  new_metadata_idx_store=None,
                                  new_y_mask=None,
                                  new_df=None,
                                  new_term_offsets=None,
                                  new_metadata_offsets=None):

        X, mX, y = self._update_X_mX_y(new_X, new_mX, new_y, new_y_mask)
        metadata_offsets, term_offsets = self._update_offsets(
            new_metadata_idx_store=new_metadata_idx_store,
            new_metadata_offsets=new_metadata_offsets,
            new_term_idx_store=new_term_idx_store,
            new_term_offsets=new_term_offsets
        )

        return OffsetCorpus(
            X=X,
            mX=mX,
            y=y,
            parsed_col=self._parsed_col,
            category_col=self._category_col,
            term_idx_store=new_term_idx_store if new_term_idx_store is not None else self._term_idx_store,
            category_idx_store=new_category_idx_store if new_category_idx_store is not None \
                else self._category_idx_store,
            metadata_idx_store=new_metadata_idx_store if new_metadata_idx_store is not None \
                else self._metadata_idx_store,
            df=self._apply_mask_to_df(new_y_mask, new_df),
            term_offsets=term_offsets,
            metadata_offsets=metadata_offsets,
            unigram_frequency_path=self._unigram_frequency_path,
        )

    def _update_offsets(
            self,
            new_metadata_idx_store,
            new_metadata_offsets,
            new_term_idx_store,
            new_term_offsets
    ):
        term_offsets = self._term_offsets if new_term_offsets is None else new_term_offsets
        metadata_offsets = self._metadata_offsets if new_metadata_offsets is None else new_metadata_offsets
        if new_term_idx_store is not None:
            term_offsets = {k: term_offsets[k] for k in new_term_idx_store.values()}

        if new_metadata_idx_store is not None:
            metadata_offsets = {k: metadata_offsets[k] for k in new_metadata_idx_store.values()}
        return metadata_offsets, term_offsets

    def use_categories_as_metadata_and_replace_terms(self):
        '''
        Returns a TermDocMatrix which is identical to self except the metadata values are now identical to the
         categories present and term-doc-matrix is now the metadata matrix.

        :return: TermDocMatrix
        '''
        new_tdm = self._make_new_term_doc_matrix(
            new_X=self._mX,
            new_mX=self._categories_to_metadata_factory(),
            new_y=self._y,
            new_term_idx_store=self._metadata_idx_store,
            new_category_idx_store=self._category_idx_store,
            new_metadata_idx_store=copy(self._category_idx_store),
            new_y_mask=self._y == self._y,
            new_term_offsets=self._metadata_offsets,
            new_metadata_offsets={k: [] for k in self.get_categories()},
        )
        return new_tdm



from typing import List, Tuple, Union, Optional
from collections import deque

from scattertext.features.featoffsets.feat_and_offset_getter import FeatAndOffsetGetter
from scattertext.Common import ENGLISH_STOP_WORDS


def window(seq, n=2):
    it = iter(seq)
    win = deque((next(it, None) for _ in range(n)), maxlen=n)
    yield list(win)
    append = win.append
    for e in it:
        append(e)
        yield list(win)


class PosStopgramFeatures(FeatAndOffsetGetter):
    def __init__(
            self,
            ngram_sizes: Optional[List[int]] = None
    ):
        self.ngram_sizes = [1, 2, 3] if ngram_sizes is None else ngram_sizes

    def get_term_offsets(self, doc):
        return []

    def get_metadata_offsets(
            self,
            doc
    ) -> List[Tuple[str, List[Union[int, List[Tuple[int, int]]]]]]:
        offset_tokens = {}
        for sent in doc.sents:
            sent_features = []
            for tok in sent:
                tok_text = tok.lower_
                if tok.lower_ not in ENGLISH_STOP_WORDS or (tok.tag_[:2] in ['VB', 'NN', 'JJ']):
                    tok_text = tok.tag_
                sent_features.append([tok.idx, tok.idx + len(tok), tok_text])
            for ngram_size in self.ngram_sizes:
                if len(sent_features) >= ngram_size:
                    for ngram in window(sent_features, ngram_size):
                        toktext = ' '.join(x[2] for x in ngram)
                        token_stats = offset_tokens.setdefault(toktext, [0, []])
                        token_stats[0] += 1
                        start, end = ngram[0][0], ngram[-1][1]
                        token_stats[1].append((start, end))
        return list(offset_tokens.items())



from collections import Counter


def make_unique(lst, sep='-'):
    seen = Counter()
    out = []
    counts = Counter(lst)
    for el in lst:
        out.append(el + ('' if counts[el] == 1 else sep + str(seen[el] + 1)))
        seen[el] += 1
    return out



import numpy as np
import pandas as pd
class Lowess(object):
    def __init__(self):
        self.points = None

    def fit(self, xdata, ydata):
        import statsmodels.api as sm
        self.model = sm.nonparametric.lowess(ydata, xdata)#, frac=1./3)
        return self

    def predict(self, x):
        return np.interp(x, self.model.T[0], self.model.T[1])

    def fit_predict(self, xdata, ydata):
        df = pd.DataFrame({
            'X': np.array(xdata),
            'Y': np.array(ydata)
        }).sort_values(by='X')

        return self.fit(df.X, df.Y).predict(xdata)



from scattertext.WhitespaceNLP import Tok, _get_pos_tag, Doc


class RobertaTokenizerWrapper:
    '''
    Encapsulates the roberta tokenizer
    '''
    def __init__(self,
                 tokenizer,
                 decoder=None,
                 entity_type=None,
                 tag_type=None):
        self.tokenizer = tokenizer
        if decoder is None:
            try:
                from text_unidecode import unidecode
            except:
                raise Exception("Please install the text_unicode package to preprocess documents. "
                                "If you'd like to bypass this step, pass a text preprocessing "
                                "(e.g., lambda x: x) function into the decode parameter of this class.")
            self.decoder = unidecode
        else:
            self.decoder = decoder
        self.entity_type = entity_type
        self.tag_type = tag_type

    def tokenize(self, doc):
        '''
        doc: str, text to be tokenized
        '''

        sents = []
        decoded_text = self.decoder(doc)
        tokens = self.tokenizer.convert_ids_to_tokens(
            self.tokenizer(decoded_text)['input_ids'],
            skip_special_tokens=True
        )

        last_idx = 0
        toks = []
        for raw_token in tokens:
            token_surface_string = raw_token
            if ord(raw_token[0]) == 288:
                token_surface_string = raw_token[1:]
            if ord(raw_token[0]) == 266: # skip new lines
                last_idx += len(raw_token)
                continue
            token_idx = decoded_text.index(token_surface_string, last_idx)
            toks.append(Tok(_get_pos_tag(token_surface_string),
                            token_surface_string.lower(),
                            raw_token.lower(),
                            ent_type='' if self.entity_type is None else self.entity_type.get(token_surface_string, ''),
                            tag='' if self.tag_type is None else self.tag_type.get(token_surface_string, ''),
                            idx=token_idx))
            last_idx = token_idx + len(token_surface_string)
            if token_surface_string in ['.', '!', '?']: # idiot's sentence splitter
                sents.append(toks)
                toks = []

        if len(toks) > 0:
            sents.append(toks)
        return Doc(sents, decoded_text)


    def get_subword_encoding_name(self):
        return 'RoBERTa'


from scattertext.diachronic.DiachronicVisualizer import DiachronicVisualizer


class BubbleDiachronicVisualization(DiachronicVisualizer):
    @staticmethod
    def visualize(display_df):
        viridis = ['#440154', '#472c7a', '#3b518b', '#2c718e', '#21908d', '#27ad81', '#5cc863', '#aadc32', '#fde725']
        import altair as alt
        color_scale = alt.Scale(
            domain=(display_df.dropna().trending.min(),
                    0,
                    display_df.dropna().trending.max()),
            range=[viridis[0], viridis[len(viridis) // 2], viridis[-1]]
        )

        return alt.Chart(display_df).mark_circle().encode(
            alt.X('variable'),
            alt.Y('term'),
            size='frequency',
            color=alt.Color('trending:Q', scale=color_scale),
        )



import unittest
import pandas as pd
from scattertext.WhitespaceNLP import whitespace_nlp

from scattertext import CorpusWithoutCategoriesFromParsedDocuments


class TestCorpusWithoutCategoriesFromParsedDocuments(unittest.TestCase):
    def test_main(self):
        doc_df = pd.DataFrame({
            'Text': [x.strip() for x in '''b a m n i b e u p
        b a s a t b e w q n
        b c a g a b e s t a
        b a g h a b e a a t
        b a h a a b e a x a t'''.split('\n')]}).assign(
            Parse=lambda df: df.Text.apply(whitespace_nlp)
        )
        corpus = CorpusWithoutCategoriesFromParsedDocuments(doc_df, parsed_col='Parse').build()



import pandas as pd

from scattertext.DataFrameCorpus import DataFrameCorpus


class CorpusDF(DataFrameCorpus):
    def __init__(self,
                 df,
                 X,
                 mX,
                 y,
                 text_col,
                 term_idx_store,
                 category_idx_store,
                 metadata_idx_store,
                 unigram_frequency_path=None):
        '''
        Parameters
        ----------
        X : csr_matrix
            term document matrix
        mX : csr_matrix
            metadata-document matrix
        y : np.array
            category index array
        term_idx_store : IndexStore
            Term indices
        category_idx_store : IndexStore
            Catgory indices
        metadata_idx_store : IndexStore
          Document metadata indices
        text_col: np.array or pd.Series
            Raw texts
        unigram_frequency_path : str or None
            Path to term frequency file.
        '''
        self._text_col = text_col
        DataFrameCorpus.__init__(self,
                                 X,
                                 mX,
                                 y,
                                 term_idx_store,
                                 category_idx_store,
                                 metadata_idx_store,
                                 df[text_col],
                                 df,
                                 unigram_frequency_path)

    def get_texts(self):
        '''
        Returns
        -------
        pd.Series, all raw documents
        '''
        return self._df[self._text_col]


    def _make_new_term_doc_matrix(self,
                                  new_X=None,
                                  new_mX=None,
                                  new_y=None,
                                  new_term_idx_store=None,
                                  new_category_idx_store=None,
                                  new_metadata_idx_store=None,
                                  new_y_mask=None,
                                  new_df=None):
        X, mX, y = self._update_X_mX_y(new_X, new_mX, new_y, new_y_mask)

        return CorpusDF(
            df=self._apply_mask_to_df(new_y_mask, new_df),
            X=X,
            mX=mX,
            y=y,
            term_idx_store=new_term_idx_store if new_term_idx_store is not None else self._term_idx_store,
            category_idx_store=new_category_idx_store if new_category_idx_store is not None else self._category_idx_store,
            metadata_idx_store=new_metadata_idx_store if new_metadata_idx_store is not None else self._metadata_idx_store,
            text_col=self._text_col,
            unigram_frequency_path=self._unigram_frequency_path
        )



from scattertext.graphs.GraphStructure import GraphStructure


class TimeStructure(GraphStructure):
    def __init__(self,
                 scatterplot_structure,
                 graph_renderer,
                 scatterplot_width=500,
                 scatterplot_height=700,
                 d3_url_struct=None,
                 protocol='http',
                 template_file_name='time_plot.html'):
        GraphStructure.__init__(self,
                                scatterplot_structure,
                                graph_renderer,
                                scatterplot_width,
                                scatterplot_height,
                                d3_url_struct,
                                protocol,
                                template_file_name)

    def _replace_html_template(self, autocomplete_css, html_template, javascript_to_insert):
        html_template = html_template.replace(
            '<!-- EXTRA LIBS -->',
            "<script src='../scattertext/scattertext/data/viz/scripts/timelines-chart.js'></script>\n<!--D3URL-->"
        )
        return GraphStructure._replace_html_template(self, autocomplete_css, html_template, javascript_to_insert)



from scipy.sparse import csc_matrix, csr_matrix
from sklearn.preprocessing import StandardScaler
import numpy as np
import pandas as pd

from scattertext.smoothing.lowess import Lowess


class Dispersion(object):
    def __init__(self, corpus=None, term_doc_mat=None, use_metadata=False, use_categories=False, tqdm=None):
        """
        From https://www.researchgate.net/publication/332120488_Analyzing_dispersion
        Stefan Th. Gries. Analyzing dispersion. April 2019. Practical handbook of corpus linguistics. Springer.

        Parts are considered documents

        :param X: term document (part) matrix
        """

        '''
        following Gries' notation, for the following example:
        b a m n i b e u p
        b a s a t b e w q n
        b c a g a b e s t a
        b a g h a b e a a t
        b a h a a b e a x a t

        (1) l = 50 (the length of the corpus in words)
        (2) n = 5 (the length of the corpus in parts)
        (3) s = (0.18, 0.2, 0.2, 0.2, 0.22) (the percentages of the n corpus part sizes)
        (4) f = 15 (the overall frequency of a in the corpus)
        (5) v = (1, 2, 3, 4, 5) (the frequencies of a in each corpus part 1-n)
        (6) p = (1/9, 2/10, 3/10, 4/10, 5 /11) (the percentages a makes up of each corpus part 1-n)
        '''
        self.corpus = None
        self.use_metadata = use_metadata
        X = term_doc_mat
        if corpus is not None:
            self.corpus = corpus
            if use_metadata:
                if use_categories is False:
                    X = corpus.get_metadata_doc_mat()
                else:
                    X = csr_matrix(corpus.get_metadata_freq_df().values.T)
            else:
                if use_categories is False:
                    X = corpus.get_term_doc_mat()
                else:
                    X = csr_matrix(corpus.get_term_freq_df().values.T)
        part_sizes = X.sum(axis=1)
        self.l = X.sum().sum()
        self.n = X.shape[0]
        self.f = X.sum(axis=0)
        self.v = X
        self.p = X.multiply(csc_matrix(1. / X.sum(axis=1)))
        self.s = part_sizes / self.l
        self.tqdm = tqdm

    def dispersion_range(self):
        """
        range: number of parts containing a = 5
        """
        return (self.v > 0).sum(axis=0).A1

    def sd_population(self):
        return np.sqrt(StandardScaler(with_mean=False).fit(self.v).var_)

    def vc(self):
        """
        Direct quote from Gries (2019)
        A maybe more useful variant of this measure is its normalized version, the variation
        coefficient (vc, see (9)); the normalization consists of dividing sdpopulation by the mean frequency
        of the element in the corpus parts f/n:
        """
        ss = StandardScaler(with_mean=False).fit(self.v)
        return np.sqrt(ss.var_) / ss.mean_

    def jullands_d(self):
        """
        Direct quote from Gries (2019)

        The version of Juilland's D that can handle differently large corpus parts is then computed
        as shown in (10). In order to accommodate the different sizes of the corpus parts, however, the
        variation coefficient is not computed using the observed frequencies v1-n (i.e. 1, 2, 3, 4, 5 in files
        1 to 5 respectively, see (5) above) but using the percentages in p1-n (i.e. how much of each corpus
        part is made up by the element in question, i.e. 1/9, 2/10, 3/10, 4/10, 5/11, see (6) above), which is what
        corrects for differently large corpus parts:
        """
        ss = StandardScaler(with_mean=False).fit(self.p)
        return 1 - (np.sqrt(ss.var_) / ss.mean_) / np.sqrt(self.n - 1)

    def rosengrens(self):
        '''
        Direct quote from Gries (2019)

        The version of Rosengren’s S that can handle differently large corpus parts is
        shown in (12). Each corpus part size’s in percent (in s) is multiplied with the
        frequencies of the element in question in each corpus part (in v1-n); of each product,
        one takes the square root, and those are summed up, that sum is squared, and divided
        by the overall frequency of the element in question in the corpus (f)'''

        vs = self.v.multiply(self.s)
        return np.power(np.sqrt(vs).sum(axis=0).A1, 2) * 1. / self.get_frequency()

    def dp(self):
        '''
        Direct quote from Gries (2019)

        Finally, Gries (2008, 2010) and the follow-up by Lijffijt and Gries (2012)
        proposed a measure called DP (for deviation of proportions), which falls between
        1-min s (for an extremely even distribution) and 1 (for an extremely clumpy
        distribution) as well as a normalized version of DP, DPnorm, which falls between 0
        and 1, which are computed as shown in (13). For DP, one computes the differences
        between how much of the element in question is in each corpus file in percent on the
        one hand and the sizes of the corpus parts in percent on the other – i.e. the differences
        between observed and expected percentages. Then, one adds up the absolute values
        of those and multiplies by 0.5; the normalization then consists of dividing this values
        by the theoretically maximum value of DP given the number of corpus parts (in a
        way reminiscent of (11)'''
        return np.sum(np.abs(self.v.multiply(1. / self.get_frequency()) - self.s),
                      axis=0).A1 / 2

    def dp_norm(self):
        return self.dp() / (1 - self.s.min())

    def kl_divergence(self):
        '''
        Direct quote from Gries (2019)
        The final measure to be discussed here is one that, as far as I can tell, has never
        been proposed as a measure of dispersion, but seems to me to be ideally suited to be
        one, namely the Kullback-Leibler (or KL-) divergence, a non-symmetric measure
        that quantifies how different one probability distribution (e.g., the distribution of
        all the occurrences of a across all corpus parts, i.e. v/f) is from another (e.g., the
        corpus part sizes s); the KL-divergence is computed as shown in (14) (with log2s of 167
        0 defined as 0):'''
        vf = self.v.multiply(1. / self.f)
        vfs = vf.multiply(1. / self.s)
        vfs.data = np.log(vfs.data) / np.log(2)
        return np.sum(vf.multiply(vfs), axis=0).A1

    def da(self):
        '''
        Metrics from Burch (2017).

        Brent Burch, Jesse Egbertb and Douglas Biber. Measuring Lexical Dispersion in Corpus Linguistics. JRDS. 2016.
        Article: https://journal.equinoxpub.com/JRDS/article/view/9480

        D_A = 1 - ((n * (n - 1))/2) * sum_{i in 0, n - 1} sum{j in i + 1, n} |v_i - v_j|/(2*mean(v))

        :return:
        '''
        n = self.n

        constant = 1. / (n * (n - 1) / 2)

        it = range(self.v.shape[1])
        if self.tqdm is not None:
            it = self.tqdm(it)

        da = []
        for word_i in it:
            v_word_id = self.v.T[word_i]
            if type(v_word_id) != np.ndarray:
                y = v_word_id.todense().A1
            else:
                y = v_word_id
            yt = np.tile(y, (n, 1))
            pairs_sum = np.sum(np.abs(yt - yt.T)) / 2
            da_score = 1 - pairs_sum * constant / (2 * y.mean())
            da.append(da_score)

        return np.array(da)

    def get_df(self, terms=None):
        if terms is None and self.corpus is not None:
            terms = self.get_names()

        freq = self.get_frequency()
        df_content = {
            'Frequency': freq,
            'Range': self.dispersion_range(),
            'SD': self.sd_population(),
            'VC': self.vc(),
            "Juilland's D": self.jullands_d(),
            "Rosengren's S": self.rosengrens(),
            'DP': self.dp(),
            'DP norm': self.dp_norm(),
            'KL-divergence': self.kl_divergence(),
        }
        if terms is None:
            df = pd.DataFrame(df_content)
        else:
            df = pd.DataFrame(df_content, index=terms)
        return df

    def get_names(self):
        return self.corpus.get_metadata() if self.use_metadata else self.corpus.get_terms()

    def get_adjusted_metric(self, metric=None, freq=None):
        '''
        Returns the difference between DA and the Lowess estimate of DP from frequency

        :param metiric: Optional[np.array], metric to analyze, defaults to DP

        :param freq: Optional[np.array], Word frequencies
        :return: np.array, frequency-adjusted metric
        '''
        if metric is None:
            metric = self.dp()
        if freq is None:
            freq = self.get_frequency()
        freq_est_metric = Lowess().fit_predict(freq, metric)
        adjusted_metric = metric - freq_est_metric
        return adjusted_metric

    def get_frequency(self):
        if len(self.f.shape) == 1:
            return self.f
        return self.f.A1



import io
import pkgutil
from collections import Counter
from re import split
from sys import version_info

import pandas as pd

from scattertext.features.FeatsFromSpacyDoc import FeatsFromSpacyDoc


class FeatsFromMoralFoundationsDictionary(FeatsFromSpacyDoc):
    def __init__(self,
                 use_lemmas=False,
                 entity_types_to_censor=set(),
                 tag_types_to_censor=set(),
                 strip_final_period=False,
                 **kwargs):
        '''
        Parameters
        ----------
        Other parameters from FeatsFromSpacyDoc.__init__
        '''
        self._lexicon_df = self._load_mfd()
        super(FeatsFromMoralFoundationsDictionary, self).__init__(use_lemmas,
                                                                  entity_types_to_censor,
                                                                  tag_types_to_censor,
                                                                  strip_final_period)

    def _load_mfd(self):
        return pd.read_csv(
            io.StringIO(pkgutil.get_data('scattertext', 'data/mfd2.0.csv').decode('utf-8'))
        ).set_index('term')

    def _analyze(self, doc):
        text_df = (pd.DataFrame(pd.Series(Counter(t for t in split(r"(\W)", doc.lower()) if t.strip())))
                   .join(self._lexicon_df)
                   .dropna()
                   .groupby('cat')
                   .sum()
                   )
        return text_df

    def get_definitions(self):
        '''
        These definitions are from https://osf.io/xakyw/

        :return: dict
        '''
        return {
            'care.virtue': '...acted with kindness, compassion, or empathy, or nurtured another person.',
            'care.vice': '...acted with cruelty, or hurt or harmed another person/animal and caused suffering.',
            'fairness.virtue': '...acted in a fair manner, promoting equality, justice, or rights.',
            'fairness.vice': '...was unfair or cheated, or caused an injustice or engaged in fraud.',
            'loyalty.virtue': '...acted with fidelity, or as a team player, or was loyal or patriotic.',
            'loyalty.vice': '...acted disloyal, betrayed someone, was disloyal, or was a traitor.',
            'authority.virtue': '...obeyed, or acted with respect for authority or tradition.',
            'authority.vice': '...disobeyed or showed disrespect, or engaged in subversion or caused chaos',
            'sanctity.virtue': '...acted in a way that was wholesome or sacred, or displayed purity or sanctity',
            'sanctity.vice': '...was depraved, degrading, impure, or unnatural.'
        }

    def get_doc_metadata(self, doc, prefix=''):
        topic_counter = Counter()
        if version_info[0] >= 3:
            doc = str(doc)
        for topic_category, score in self._analyze(doc).to_dict()[0].items():
            topic_counter[prefix + topic_category] = int(score)
        return topic_counter

    def has_metadata_term_list(self):
        return True

    def get_top_model_term_lists(self):
        return self._lexicon_df.reset_index().groupby('cat')['term'].apply(list).to_dict()



import numpy as np

from scattertext.CSRMatrixTools import CSRMatrixFactory
from scattertext.ParsedCorpus import ParsedCorpus
from scattertext.features.FeatsFromSpacyDoc import FeatsFromSpacyDoc
from scattertext.indexstore.IndexStore import IndexStore


class CorpusFromParsedDocuments(object):
    def __init__(self,
                 df,
                 category_col,
                 parsed_col,
                 feats_from_spacy_doc=FeatsFromSpacyDoc()):

        '''
        Parameters
        ----------
        df : pd.DataFrame
         contains category_col, and parse_col, were parsed col is entirely spacy docs
        category_col : str
            name of category column in convention_df
        parsed_col : str
            name of spacy parsed column in convention_df
        feats_from_spacy_doc : FeatsFromSpacyDoc
        '''
        self._df = df.reset_index()
        self._category_col = category_col
        self._parsed_col = parsed_col
        self._category_idx_store = IndexStore()
        self._X_factory = CSRMatrixFactory()
        self._mX_factory = CSRMatrixFactory()
        self._term_idx_store = IndexStore()
        self._metadata_idx_store = IndexStore()
        self._feats_from_spacy_doc = feats_from_spacy_doc

    def build(self, show_progress=False):
        '''Constructs the term doc matrix.

        Returns
        -------
        scattertext.ParsedCorpus.ParsedCorpus
        '''
        y = self._get_y_and_populate_category_idx_store(self._df[self._category_col])
        if show_progress is True:
            self._df.progress_apply(self._add_to_x_factory, axis=1)
        else:
            self._df.apply(self._add_to_x_factory, axis=1)
        self._mX = self._mX_factory.set_last_row_idx(len(y) - 1).get_csr_matrix()
        return ParsedCorpus(
            df=self._df,
            X=self._X_factory.set_last_row_idx(len(y) - 1).get_csr_matrix(),
            mX=self._mX_factory.set_last_row_idx(len(y) - 1).get_csr_matrix(),
            y=y,
            term_idx_store=self._term_idx_store,
            category_idx_store=self._category_idx_store,
            metadata_idx_store=self._metadata_idx_store,
            parsed_col=self._parsed_col,
            category_col=self._category_col
        )

    def _get_y_and_populate_category_idx_store(self, categories):
        return np.array(categories.apply(self._category_idx_store.getidx))

    def _add_to_x_factory(self, row):
        parsed_text = row[self._parsed_col]
        for term, count in self._feats_from_spacy_doc.get_feats(parsed_text).items():
            term_idx = self._term_idx_store.getidx(term)
            self._X_factory[row.name, term_idx] = count
        for meta, val in self._feats_from_spacy_doc.get_doc_metadata(parsed_text).items():
            meta_idx = self._metadata_idx_store.getidx(meta)
            self._mX_factory[row.name, meta_idx] = val



import unittest
from turtle import st

import pandas as pd
import numpy as np
from scattertext.WhitespaceNLP import whitespace_nlp

from scattertext import Dispersion, CorpusWithoutCategoriesFromParsedDocuments


class TestDispersion(unittest.TestCase):
    def test_main(self):
        doc_df = pd.DataFrame({
            'Text': [x.strip() for x in '''b a m n i b e u p
        b a s a t b e w q n
        b c a g a b e s t a
        b a g h a b e a a t
        b a h a a b e a x a t'''.split('\n')]}).assign(
            Parse=lambda df: df.Text.apply(whitespace_nlp),
        )
        corpus = CorpusWithoutCategoriesFromParsedDocuments(
            doc_df, parsed_col='Parse'
        ).build(
        ).get_unigram_corpus()
        dispersion_df = Dispersion(corpus).get_df()
        dispersion_df = dispersion_df.loc[['b', 'i', 'q', 'x']]
        assert list(dispersion_df['Range'].values) == [5, 1, 1, 1]
        assert list(dispersion_df['SD'].values) == [0, 0.4, 0.4, 0.4]
        assert list(dispersion_df['VC'].values) == [0, 2, 2, 2]
        np.testing.assert_almost_equal(
            list(dispersion_df["Juilland's D"].values),
            np.array([0.968, 0, 0, 0]),
            decimal=3
        )
        np.testing.assert_almost_equal(
            list(dispersion_df["Rosengren's S"].values),
            np.array([0.999, 0.18, 0.2, 0.22]),
            decimal=3

        )
        np.testing.assert_almost_equal(
            list(dispersion_df["DP"].values),
            np.array([0.02, 0.82, 0.8, 0.78]),
            decimal=3
        )
        np.testing.assert_almost_equal(
            list(dispersion_df["DP norm"].values),
            np.array([0.024, 1, 0.976, 0.951]),
            decimal=3
        )
        np.testing.assert_almost_equal(
            list(dispersion_df["KL-divergence"].values),
            np.array([0.003, 2.474, 2.322, 2.184]),
            decimal=3
        )



import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression

from scattertext.diachronic.BubbleDiachronicVisualization import BubbleDiachronicVisualization
from scattertext.diachronic.DiachronicVisualizer import DiachronicVisualizer


class DiachronicTermMiner(object):
    def __init__(self,
                 corpus,
                 num_terms=40,
                 start_category=None,
                 timesteps_to_lag=5,
                 seasonality_column=None,
                 model=None):
        self.corpus_ = corpus
        self.num_terms_ = num_terms
        self.sorted_categores_ = np.array(sorted(corpus.get_categories()))
        if timesteps_to_lag > len(self.sorted_categores_):
            raise Exception("Too many timesteps to lag. There are only %s categories" %
                            len(self.sorted_categores_))
        self.timesteps_to_lag_ = timesteps_to_lag
        if start_category is None:
            self.start_category_ = sorted(corpus.get_categories())[timesteps_to_lag]
        else:
            if start_category not in corpus.get_categories():
                raise Exception("start_category %s needs to be a category in corpus" % (start_category))
            if start_category in self.sorted_categores_[:timesteps_to_lag]:
                raise Exception("start_category %s should have at least %s categories before it"
                                % (start_category, timesteps_to_lag))
            self.start_category_ = start_category
        if seasonality_column:
            if seasonality_column not in self.corpus_.get_df().columns:
                raise Exception("seasonality_column should be none or a column in the source dataframe")
        self.seasonality_column_ = seasonality_column
        self.model_ = model

    def get_display_dataframe(self):
        '''
        Gets list of terms to display that have some interesting diachronic variation.

        Returns
        -------
        pd.DataFrame

        e.g.,
                   term variable  frequency  trending
        2            in   200310        1.0  0.000000
        19          for   200310        1.0  0.000000
        20           to   200311        1.0  0.000000
        '''
        X = self.corpus_.get_term_doc_mat()
        categories = pd.Series(self.corpus_.get_category_ids())
        cat_ar = np.array(self.corpus_.get_categories())
        cat_idx_sort = np.argsort(cat_ar)
        if self.seasonality_column_:
            seasonality_ar = np.array(self.corpus_.get_df()[self.seasonality_column_])

        terms = self.corpus_.get_terms()
        category_idx_store = self.corpus_.get_category_index_store()

        data = {}
        seasondata = {}
        for i, cati in enumerate(cat_idx_sort):
            cat = cat_ar[cati]
            if cat >= self.start_category_ and i > self.timesteps_to_lag_:
                neg_cats = self.sorted_categores_[i - self.timesteps_to_lag_:i]
                neg_mask = categories.isin(category_idx_store.getidxstrictbatch(neg_cats)).values
                scores = self._regress_terms(X, cat, categories, category_idx_store, neg_mask, terms)
                data[cat] = scores
                if self.seasonality_column_:
                    neg_cats = set(
                        categories[(seasonality_ar == seasonality_ar[cati]) & (categories != categories[cati])])
                    neg_mask = categories.isin(neg_cats).values
                    scores = self._regress_terms(X, cat, categories, category_idx_store, neg_mask, terms)
                    seasondata[cat] = scores
        coefs = pd.DataFrame(data)
        pos_coefs = (coefs.apply(lambda x: (x > 0) * x, axis=1)
                     .sum(axis=1)
                     .sort_values(ascending=False))

        term_cat_counts = self.corpus_.get_term_freq_df('')[coefs.columns]

        def dense_percentile(x):
            # ranks = rankdata(x, 'dense')
            return pd.Series(x / x.max(), index=x.index)

        rank_df = pd.DataFrame({'coefr': dense_percentile(pos_coefs),
                                'freqr': dense_percentile(term_cat_counts.max(axis=1)),
                                'coef': pos_coefs,
                                'freq': term_cat_counts.max(axis=1)})
        if self.seasonality_column_:
            seasoncoefs = (pd.DataFrame(seasondata).sum(axis=1))
            rank_df['seasoncoefr'] = dense_percentile(
                seasoncoefs.sort_values(ascending=False) + np.abs(seasoncoefs.min()))
            weights = [2, 1, 1]
            vals = ['freqr', 'coefr', 'seasoncoefr']

            def gethmean(x):
                if min(x[vals]) == 0:
                    return 0
                return sum(weights) * 1. / sum([weights[i] / x[val] for i, val in enumerate(vals)])

            rank_df['hmean'] = rank_df.apply(gethmean, axis=1)
        else:
            beta = 0.5
            rank_df['hmean'] = (rank_df
                                .apply(lambda x: 0 if min(x) == 0 else
            (1 + beta ** 2) * (x.coefr * x.freqr) / ((beta ** 2 * x.coefr) + x.freqr),
                                       axis=1))
        rank_df = rank_df.sort_values(by='hmean', ascending=False)
        display_df = pd.merge(
            (term_cat_counts
             .loc[rank_df.iloc[:self.num_terms_].index]
             .reset_index()
             .melt(id_vars=['index'])
             .rename(columns={'index': 'term', 'value': 'frequency'})),
            (coefs.loc[rank_df.iloc[:self.num_terms_].index]
             .reset_index()
             .melt(id_vars=['index'])
             .rename(columns={'index': 'term', 'value': 'trending'})),
            on=['term', 'variable']
		)
        display_df[display_df['frequency'] == 0] = np.nan
        display_df = display_df.dropna()
        return display_df[display_df.term.isin(rank_df.index)]

    def visualize(self, visualizer=BubbleDiachronicVisualization):
        assert issubclass(visualizer, DiachronicVisualizer)
        return visualizer.visualize(self.get_display_dataframe())

    def _regress_terms(self, X, cat, categories, category_idx_store, neg_mask, terms):
        pos_mask = categories.isin(category_idx_store.getidxstrictbatch([cat])).values
        catX = X[neg_mask | pos_mask, :]
        catY = np.zeros(catX.shape[0]).astype(bool)
        catY[pos_mask[neg_mask | pos_mask]] = True
        scores = (pd.Series(self._get_model().fit(catX, catY).coef_[0], index=terms)
                  .sort_values(ascending=False))
        return scores

    def _get_model(self):
        return self.model_ if self.model_ is not None else LogisticRegression(penalty='l2')



from collections import Counter

from scattertext.features.FeatsFromSpacyDoc import FeatsFromSpacyDoc


class SpacyEntities(FeatsFromSpacyDoc):
    def __init__(self,
                 use_lemmas=False,
                 entity_types_to_censor=set(),
                 entity_types_to_use=None,
                 tag_types_to_censor=set(),
                 strip_final_period=False):
        self._entity_types_to_use = entity_types_to_use
        FeatsFromSpacyDoc.__init__(self, use_lemmas, entity_types_to_censor,
                                   tag_types_to_censor, strip_final_period)

    def get_feats(self, doc):
        return Counter([
            ' '.join(str(ent).split()).lower()
            for ent
            in doc.ents
            if ((self._entity_types_to_use is None
                 or ent.label_ in self._entity_types_to_use)
                and (ent.label_ not in self._entity_types_to_censor))
        ])



from collections import Counter
from scattertext.features.FeatsFromSpacyDoc import FeatsFromSpacyDoc


class PyatePhrases(FeatsFromSpacyDoc):
    def __init__(self, extractor=None, **args):
        import pyate
        self._extractor = pyate.combo_basic if extractor is None else extractor
        FeatsFromSpacyDoc.__init__(self, **args)

    def get_feats(self, doc):
        return Counter(self._extractor(str(doc)).to_dict())




import numpy as np

class DeltaJSDivergence(object):
    def __init__(self, pi1=0.5, pi2=0.5):
        assert pi1 + pi2 == 1
        self.pi1 = pi1
        self.pi2 = pi2

    def get_scores(self, a, b):
        # via https://arxiv.org/pdf/2008.02250.pdf eqn 1
        p1 = 0.001 + a / np.sum(a)
        p2 = 0.001 + b / np.sum(b)
        pi1, pi2 = self.pi1, self.pi2
        m = pi1 * p1 + pi2 * p2
        def lg(x): return np.log(x) / np.log(2)

        return m * lg(1 / m) - (pi1 * p2 * lg(1 / p1) + pi2 * p2 * lg(1 / p2))

    def get_default_score(self):
        return 0

    def get_name(self):
        return 'JS Divergence Shift'


class CoefficientBase:
    def __init__(self, use_non_text=False):
        self.use_non_text = use_non_text

    def _get_tdm(self, corpus):
        return corpus.get_metadata_doc_mat() if self.use_non_text else corpus.get_term_doc_mat()

    def _get_terms(self, corpus):
        return corpus.get_metadata() if self.use_non_text else corpus.get_terms()


import pkgutil

from scattertext.graphs.ComponentDiGraphHTMLRenderer import GraphRenderer
from scattertext.viz.BasicHTMLFromScatterplotStructure import D3URLs, ExternalJSUtilts, PackedDataUtils

GRAPH_VIZ_FILE_NAME = 'graph_plot.html'


class GraphStructure(object):
    def __init__(self,
                 scatterplot_structure,
                 graph_renderer,
                 scatterplot_width=500,
                 scatterplot_height=700,
                 d3_url_struct=None,
                 protocol='http',
                 template_file_name=None):
        ''',
        Parameters
        ----------
        scatterplot_structure: ScatterplotStructure
        graph_renderer: GraphRenderer
        scatterplot_width: int
        scatterplot_height: int
        d3_url_struct: D3URLs
        protocol: str
            http or https
        template_file_name: file name to use as template
        '''
        self.graph_renderer = graph_renderer
        self.scatterplot_structure = scatterplot_structure
        self.d3_url_struct = d3_url_struct if d3_url_struct else D3URLs()
        ExternalJSUtilts.ensure_valid_protocol(protocol)
        self.protocol = protocol

        self.scatterplot_width = scatterplot_width
        self.scatterplot_height = scatterplot_height

        self.template_file_name = GRAPH_VIZ_FILE_NAME if template_file_name is None else template_file_name

    def to_html(self):
        '''
        Returns
        -------
        str, the html file representation

        '''
        javascript_to_insert = self._get_javascript_to_insert()
        autocomplete_css = PackedDataUtils.full_content_of_default_autocomplete_css()
        html_template = self._get_html_template()
        html_content = self._replace_html_template(autocomplete_css, html_template, javascript_to_insert)
        return html_content

    def _get_javascript_to_insert(self):
        return '\n'.join([
            PackedDataUtils.full_content_of_javascript_files(),
            self.scatterplot_structure._visualization_data.to_javascript(),
            self.scatterplot_structure.get_js_to_call_build_scatterplot(),
            PackedDataUtils.javascript_post_build_viz('termSearch', 'plotInterface'),
            self.graph_renderer.get_javascript()
        ])

    def _replace_html_template(self, autocomplete_css, html_template, javascript_to_insert):
        return (html_template
                .replace('/***AUTOCOMPLETE CSS***/', autocomplete_css, 1)
                .replace('<!-- INSERT SCRIPT -->', javascript_to_insert, 1)
                .replace('<!--D3URL-->', self.d3_url_struct.get_d3_url(), 1)
                .replace('<!-- INSERT GRAPH -->', self.graph_renderer.get_graph())
                .replace('<!--D3SCALECHROMATIC-->', self.d3_url_struct.get_d3_scale_chromatic_url())
                .replace('<!--USEZOOM-->', self.get_zoom_script_import())
                .replace('<!--FONTIMPORT-->', self.get_font_import())
                .replace('http://', self.protocol + '://')
                .replace('{width}', str(self.scatterplot_width))
                .replace('{height}', str(self.scatterplot_height))
                .replace('{cellheight}', str(int(self.scatterplot_height * (6/12))))
                )

    def get_font_import(self):
        return '''<link href="https://fonts.googleapis.com/css?family=IBM+Plex+Sans&display=swap" rel="stylesheet">'''

    def get_zoom_script_import(self):
        return '<script src="https://cdn.jsdelivr.net/npm/svg-pan-zoom@3.6.0/dist/svg-pan-zoom.min.js"></script>'

    def _get_html_template(self):
        return PackedDataUtils.get_packaged_html_template_content(self.template_file_name)




import json

from scattertext.Common import DEFAULT_DIV_ID, DEFAULT_D3_AXIS_VALUE_FORMAT


class InvalidProtocolException(Exception):
    pass


class ScatterplotStructure(object):
    def __init__(
            self,
            visualization_data,
            width_in_pixels=None,
            height_in_pixels=None,
            max_snippets=None,
            color=None,
            grey_zero_scores=False,
            sort_by_dist=True,
            reverse_sort_scores_for_not_category=True,
            use_full_doc=False,
            asian_mode=False,
            match_full_line=False,
            use_non_text_features=False,
            show_characteristic=True,
            word_vec_use_p_vals=False,
            max_p_val=0.1,
            save_svg_button=False,
            p_value_colors=False,
            x_label=None,
            y_label=None,
            full_data=None,
            show_top_terms=True,
            show_neutral=False,
            get_tooltip_content=None,
            x_axis_values=None,
            y_axis_values=None,
            color_func=None,
            show_axes=True,
            horizontal_line_y_position=None,
            vertical_line_x_position=None,
            show_extra=False,
            do_censor_points=True,
            center_label_over_points=False,
            x_axis_labels=None,
            y_axis_labels=None,
            topic_model_preview_size=10,
            vertical_lines=None,
            unified_context=False,
            show_category_headings=True,
            highlight_selected_category=False,
            show_cross_axes=True,
            div_name=DEFAULT_DIV_ID,
            alternative_term_func=None,
            include_all_contexts=False,
            show_axes_and_cross_hairs=False,
            x_axis_values_format=None,
            y_axis_values_format=None,
            max_overlapping=-1,
            show_corpus_stats=True,
            sort_doc_labels_by_name=False,
            always_jump=True,
            show_diagonal=False,
            enable_term_category_description=True,
            use_global_scale=False,
            get_custom_term_html=None,
            header_names=None,
            header_sorting_algos=None,
            ignore_categories=False,
            background_labels=None,
            label_priority_column=None,
            text_color_column=None,
            suppress_text_column=None,
            censor_point_column=None,
            background_color=None,
            right_order_column=None,
            subword_encoding=None,
            top_terms_length=14,
            top_terms_left_buffer=0,
            get_column_header_html=None,
            term_word="Term"
    ):
        '''

        Parameters
        ----------
        visualization_data : VizDataAdapter
            From ScatterChart or a descendant
        width_in_pixels : int, optional
            width of viz in pixels, if None, default to 1000
        height_in_pixels : int, optional
            height of viz in pixels, if None, default to 600
        max_snippets : int, optional
            max snippets to snow when term is clicked, Defaults to show all
        color : str, optional
          d3 color scheme
        grey_zero_scores : bool, optional
          If True, color points with zero-scores a light shade of grey.  False by default.
        sort_by_dist : bool, optional
            sort by distance or score, default True
        reverse_sort_scores_for_not_category : bool, optional
            If using a custom score, score the not-category class by
            lowest-score-as-most-predictive. Turn this off for word vectory
            or topic similarity. Default True.
        use_full_doc : bool, optional
            use full document instead of sentence in snippeting
        use_non_text_features : bool, default False
            Show non-bag-of-words features (e.g., Empath) instaed of text.  False by default.
        show_characteristic: bool, default True
            Show characteristic terms on the far left-hand side of the visualization
        word_vec_use_p_vals: bool, default False
            Use category-associated p-values to determine which terms to give word
            vec scores.
        max_p_val : float, default = 0.1
            Max p-val to use find set of terms for similarity calculation, if
            word_vec_use_p_vals is True.
        save_svg_button : bool, default False
            Add a save as SVG button to the page.
        p_value_colors : bool, default False
          Color points differently if p val is above 1-max_p_val, below max_p_val, or
           in between.
        x_label : str, default None
            If present, use as the x-axis label
        y_label : str, default None
            If present, use as the y-axis label
        full_data : str, default None
            Data used to create chart. By default "getDataAndInfo()".
        show_top_terms : bool, default True
            Show the top terms sidebar
        show_neutral : bool, default False
            Show a third column for matches in neutral documents
        get_tooltip_content : str, default None
            Javascript function to control content of tooltip.  Function takes a parameter
            which is a dictionary entry produced by `ScatterChartExplorer.to_dict` and
            returns a string.
        x_axis_values : list, default None
            Numeric value-labels to show on x-axis which correspond to original x-values.
        y_axis_values : list, default None
            Numeric Value-labels to show on y-axis which correspond to original y-values.
        color_func : str, default None
            Javascript function to control color of a point.  Function takes a parameter
            which is a dictionary entry produced by `ScatterChartExplorer.to_dict` and
            returns a string.
        show_axes : bool, default True
            Show x and y axes
        horizontal_line_y_position : float, default None
            If x and y axes markers are shown, the position of the horizontal axis marker
        vertical_line_x_position : float, default None
            If x and y axes markers are shown, the position of the vertical axis marker
        show_extra : bool, default False
            Show extra fourth column
        do_censor_points  : bool, default True
            Don't label over dots
        center_label_over_points : bool, default False
            Only put labels centered over point
        x_axis_labels: list, default None
            List of string value-labels to show at evenly spaced intervals on the x-axis.
            Low, medium, high are defaults. This relies on d3's ticks function, which can
            behave unpredictable. Caveat usor.
        y_axis_labels : list, default None
            List of string value-labels to show at evenly spaced intervals on the y-axis.
            Low, medium, high are defaults.  This relies on d3's ticks function, which can
            behave unpredictable. Caveat usor.
        topic_model_preview_size : int, default 10
            If topic models are being visualized, show the first topic_model_preview_size topics
            in a preview.
        vertical_lines: list, default None
            List of scaled points along the x-axis to draw horizontal lines
        unified_context: bool, default False
            Display all context in a single column.
        show_category_headings: bool, default True
            If unified_context, should we show the category headings?
        highlight_selected_category: bool, default False
            Highlight selected category in unified view
        show_cross_axes: bool, default True
            If show_axes is False, do we show cross-axes?
        div_name: str, default DEFAULT_DIV_ID
            Div which holds scatterplot
        alternative_term_func: str, default None
            Javascript function which take a term JSON object and returns a bool.  If the return value is true,
            execute standard term click pipeline. Ex.: `'(function(termDict) {return true;})'`.
        include_all_contexts: bool, default False
            Include all contexts, even non-matching ones, in interface
        show_axes_and_cross_hairs: bool, default False
            Show both cross-axes and peripheral scales.
        x_axis_values_format: str, default None
            d3 format string for x-axis values (see https://github.com/d3/d3-format)
        y_axis_values_format: str, default None
            d3 format string for y-axis values (see https://github.com/d3/d3-format)
        max_overlapping: int, default -1
            Maximum number of overlapping terms to output.  Show all if -1 (default)
        show_corpus_stats: bool, default True
            Populate corpus stats div
        sort_doc_labels_by_name: bool, default False
            If unified document labels, sort the labels by name instead of value.
        always_jump: bool, default True
            Always jump to term contexts if a term is clicked.
        show_diagonal: bool, default False
            Show a diagonal line from the lower-left to the upper-right of the chart
        use_global_scale: bool, default False
            Use the same scaling for the x and y axis. Without this show_diagonal may not
            make sense.
        enable_term_category_description: bool, default True
            List term/metadata statistics under category
        get_custom_term_html: str, default None
            Javascript function which takes term info and outputs custom term HTML
        header_names: Dict[str, str], default None
            Dictionary giving names of term lists shown to the right of the plot. Valid keys are
            upper, lower and right.
        header_sorting_algos: Dict[str, str], default None
            Dictionary giving javascript sorting algorithms for panes. Valid keys are upper, lower
            and right. Value is a JS function which takes the "data" object.
        ignore_categories: bool, default False
            Signals plot should ignore category names.
        background_labels: List[Dict]: default None
            List of [{"Text": "Label", "X": xpos, "Y": ypos}, ...] to be background labels on plot
        label_priority_column : str, default None
            Column in term_metadata_df; larger values in the column indicate a term should be labeled first
        text_color_column: str, default None
            Column in term_metadata_df which supplies term colors
        suppress_text_column: str, default None
            Column in term_metadata_df which is of boolean value. Indicates if a term should be labeled.
        censor_point_column : str, default None
            Should we prevent labels from being drawn over a point?
        right_order_column : str, default None
            Order for right column ("characteristic" by default); largest first
        subword_encoding : str, default None
            Type of subword encoding to use, None if none, currently supports "RoBERTa"
        background_color: str, default None
            Color to set document.body's background
        sentence_piece: bool, default False
            Use sentence piece conventions from to search for terms in JS
        top_terms_length: int, default 14
            Number of top terms to show for left-hand side lists
        top_terms_lefft_buffer: int, default 0
            Number of pixels left to shift top terms list
        get_column_header_html : str, default None
            Javascript function to return html over each column. Matches header
            (Column Name, occurrences per 25k, occs, # occs * 1000/num docs, term info)
        term_word: str, default "Term"
            Word for term in visualization
        '''
        self._visualization_data = visualization_data
        self._width_in_pixels = width_in_pixels if width_in_pixels is not None else 1000
        self._height_in_pixels = height_in_pixels if height_in_pixels is not None else 600
        self._max_snippets = max_snippets
        self._color = color
        self._sort_by_dist = sort_by_dist
        self._use_full_doc = use_full_doc
        self._asian_mode = asian_mode
        self._match_full_line = match_full_line
        self._grey_zero_scores = grey_zero_scores
        self._use_non_text_features = use_non_text_features
        self._show_characteristic = show_characteristic
        self._word_vec_use_p_vals = word_vec_use_p_vals
        self._max_p_val = max_p_val
        self._save_svg_button = save_svg_button
        self._reverse_sort_scores_for_not_category = reverse_sort_scores_for_not_category
        self._p_value_colors = p_value_colors
        self._x_label = x_label
        self._y_label = y_label
        self._full_data = full_data
        self._show_top_terms = show_top_terms
        self._show_neutral = show_neutral
        self._get_tooltip_content = get_tooltip_content
        self._x_axis_values = x_axis_values
        self._y_axis_values = y_axis_values
        self._x_axis_labels = x_axis_labels
        self._y_axis_labels = y_axis_labels
        self._color_func = color_func
        self._show_axes = show_axes
        self._horizontal_line_y_position = horizontal_line_y_position
        self._vertical_line_x_position = vertical_line_x_position
        self._show_extra = show_extra
        self._do_censor_points = do_censor_points
        self._center_label_over_points = center_label_over_points
        self._topic_model_preview_size = topic_model_preview_size
        self._vertical_lines = vertical_lines
        self._unified_context = unified_context
        self._show_category_headings = show_category_headings
        self._highlight_selected_category = highlight_selected_category
        self._show_cross_axes = show_cross_axes
        self._div_name = div_name
        self._alternative_term_func = alternative_term_func
        self._include_all_contexts = include_all_contexts
        self._show_axes_and_cross_hairs = show_axes_and_cross_hairs
        self._x_axis_values_format = x_axis_values_format
        self._y_axis_values_format = y_axis_values_format
        self._max_overlapping = max_overlapping
        self._show_corpus_stats = show_corpus_stats
        self._sort_doc_labels_by_name = sort_doc_labels_by_name
        self._always_jump = always_jump
        self._show_diagonal = show_diagonal
        self._use_global_scale = use_global_scale
        self._enable_term_category_description = enable_term_category_description
        self._get_custom_term_html = get_custom_term_html
        self._header_names = header_names
        self._header_sorting_algos = header_sorting_algos
        self._ignore_categories = ignore_categories
        self._background_labels = background_labels
        self._label_priority_column = label_priority_column
        self._text_color_column = text_color_column
        self._suppress_label_column = suppress_text_column
        self._censor_point_column = censor_point_column
        self._right_order_column = right_order_column
        self._background_color = background_color
        self._subword_encoding = subword_encoding
        self._top_terms_length = top_terms_length
        self._top_terms_left_buffer = top_terms_left_buffer
        self._get_column_header_html = get_column_header_html
        self._term_word = term_word

    def call_build_visualization_in_javascript(self):
        def js_default_value(x, default='undefined'):
            return default if x is None else str(x)

        def js_default_string(x, default_string=None):
            if x is not None:
                return json.dumps(str(x))
            if default_string is None:
                return 'undefined'
            return json.dumps(default_string)

        def js_default_value_to_null(x):
            return 'null' if x is None else str(x)

        def json_or_null(x):
            return 'null' if x is None else json.dumps(x)

        def js_bool(x):
            return 'true' if x else 'false'

        def js_float(x):
            return str(float(x))

        def js_int(x):
            return str(int(x))

        def js_default_full_data(full_data):
            return full_data if full_data is not None else "getDataAndInfo()"

        def json_with_jsvalue_or_null(x):
            if x is None:
                return 'null'
            to_ret = '{'
            first = True
            for key, val in sorted(x.items()):
                if not first: to_ret += ', '
                to_ret += '"%s": %s' % (key, val)
                first = False
            to_ret += '}'
            return to_ret

        arguments = [
            js_default_value(self._width_in_pixels),
            js_default_value(self._height_in_pixels),
            js_default_value_to_null(self._max_snippets),
            js_default_value_to_null(self._color),
            js_bool(self._sort_by_dist),
            js_bool(self._use_full_doc),
            js_bool(self._grey_zero_scores),
            js_bool(self._asian_mode),
            js_bool(self._use_non_text_features),
            js_bool(self._show_characteristic),
            js_bool(self._word_vec_use_p_vals),
            js_bool(self._save_svg_button),
            js_bool(self._reverse_sort_scores_for_not_category),
            js_float(self._max_p_val),
            js_bool(self._p_value_colors),
            js_default_string(self._x_label),
            js_default_string(self._y_label),
            js_default_full_data(self._full_data),
            js_bool(self._show_top_terms),
            js_bool(self._show_neutral),
            js_default_value_to_null(self._get_tooltip_content),
            js_default_value_to_null(self._x_axis_values),
            js_default_value_to_null(self._y_axis_values),
            js_default_value_to_null(self._color_func),
            js_bool(self._show_axes),
            js_bool(self._show_extra),
            js_bool(self._do_censor_points),
            js_bool(self._center_label_over_points),
            json_or_null(self._x_axis_labels),
            json_or_null(self._y_axis_labels),
            js_default_value(self._topic_model_preview_size),
            json_or_null(self._vertical_lines),
            js_default_value_to_null(self._horizontal_line_y_position),
            js_default_value_to_null(self._vertical_line_x_position),
            js_bool(self._unified_context),
            js_bool(self._show_category_headings),
            js_bool(self._show_cross_axes),
            js_default_string(self._div_name),
            js_default_value_to_null(self._alternative_term_func),
            js_bool(self._include_all_contexts),
            js_bool(self._show_axes_and_cross_hairs),
            js_default_string(self._x_axis_values_format, DEFAULT_D3_AXIS_VALUE_FORMAT),
            js_default_string(self._y_axis_values_format, DEFAULT_D3_AXIS_VALUE_FORMAT),
            js_bool(self._match_full_line),
            js_int(self._max_overlapping),
            js_bool(self._show_corpus_stats),
            js_bool(self._sort_doc_labels_by_name),
            js_bool(self._always_jump),
            js_bool(self._highlight_selected_category),
            js_bool(self._show_diagonal),
            js_bool(self._use_global_scale),
            js_bool(self._enable_term_category_description),
            js_default_value_to_null(self._get_custom_term_html),
            json_or_null(self._header_names),
            json_with_jsvalue_or_null(self._header_sorting_algos),
            js_bool(self._ignore_categories),
            json_or_null(self._background_labels),
            js_default_string(self._label_priority_column),
            js_default_string(self._text_color_column),
            js_default_string(self._suppress_label_column),
            js_default_string(self._background_color),
            js_default_string(self._censor_point_column),
            js_default_string(self._right_order_column),
            js_default_string(self._subword_encoding),
            js_default_value(self._top_terms_length, '14'),
            js_default_value(self._top_terms_left_buffer, '0'),
            js_default_value_to_null(self._get_column_header_html),
            js_default_string(self._term_word)
        ]
        return 'buildViz(' + ',\n'.join(arguments) + ');\n'

    def get_js_to_call_build_scatterplot(self, object_name='plotInterface'):
        return object_name + ' = ' + self.call_build_visualization_in_javascript()

    def get_js_to_call_build_scatterplot_with_a_function(self, object_name='plotInterface', function_name=None):
        if function_name is None:
            function_name = 'build' + object_name
        function_text = ('function ' + function_name + '() { return '
                         + self.call_build_visualization_in_javascript() + ';}')
        return function_text + '\n\n' + object_name + ' = ' + function_name + '();'

    def get_js_reset_function(self, values_to_set, functions_to_reset, reset_function_name='reset'):
        '''

        :param functions_to_reset: List[str]
        :param values_to_set: List[str]
        :param reset_function_name: str, default = rest
        :return: str
        '''
        return ('function ' + reset_function_name + '() {'
                + "document.querySelectorAll('.scattertext').forEach(element=>element.innerHTML=null);\n"
                + "document.querySelectorAll('#d3-div-1-corpus-stats').forEach(element=>element.innerHTML=null);\n"
                + ' '.join([value + ' = ' + function_name + '();'
                            for value, function_name
                            in zip(values_to_set, functions_to_reset)])
                + '}')



"""
Approach adapted from https://www.cis.upenn.edu/~ungar/sentiment/machine_learning_SA.pdf
"""
import logging

import numpy as np
import pandas as pd

from scattertext.continuous.coefficientbase import CoefficientBase


def ungar_transform(tdm):
    pw = tdm.sum(axis=0).A1 / tdm.sum()  # get prob(word)
    tdmd = tdm.todense()
    tdmdpw = tdmd.T
    #tdmdpw = np.multiply(tdmd, pw) # use word probabilities
    tdmdpwln = (tdmdpw / tdmdpw.sum(axis=1)).T # length normalized
    print(tdmdpwln.shape, tdm.shape)
    X = 2 * np.sqrt(tdmdpwln + 3. / 8)  # Anscombe transformed
    return X.T


class UngarCoefficients(CoefficientBase):
    def get_coefficient_df(self, corpus, document_scores):
        from statsmodels.regression.linear_model import OLS
        '''

        :param corpus: TermDocMatrix, should just have unigrams
        :param document_scores: np.array, continuous value for each document score
        :return: pd.DataFrame
        '''
        assert document_scores.shape == (corpus.get_num_docs(),)
        if any(' ' in t for t in self._get_terms(corpus)):
            logging.warning('UngerCoefficients is currently designed for only unigram terms. '
                            'Run corpus.get_unigram_corpus() before using this.')

        X = ungar_transform(self._get_tdm(corpus))
        model = OLS(document_scores, X.T).fit()
        df = pd.DataFrame({
            'Word': self._get_terms(corpus),
            'Beta': model.params,
            'Tstat': model.tvalues,
            'Frequency': corpus.get_term_doc_mat().sum(axis=0)[0].A1
        }).set_index('Word')

        return df




from abc import ABC, abstractmethod
from collections import Counter
from functools import reduce
from re import split
from sys import version_info

import pandas as pd
from flashtext import KeywordProcessor

from scattertext.ScatterChart import check_topic_model_string_format
from scattertext.features.FeatsFromSpacyDoc import FeatsFromSpacyDoc


class FeatsFromTopicModelBase(ABC):
    def __init__(self, topic_model):
        self._topic_model = topic_model
        self._lexicon_df = self._get_lexicon_df_from_topic_model(topic_model)

    def _get_lexicon_df_from_topic_model(self, topic_model):
        return (pd.DataFrame(pd.Series(topic_model)
                             .apply(pd.Series)
                             .reset_index())
                .melt(id_vars=['index'])
                [['index', 'value']]
                .rename(columns={'index': 'cat', 'value': 'term'})
                .set_index('term'))

    def _analyze(self, doc):
        text_df = (pd.DataFrame(pd.Series(self._get_terms_from_doc(doc)))
                   .join(self._lexicon_df)
                   .dropna()
                   .groupby('cat')
                   .sum())
        return text_df

    def get_doc_metadata(self, doc, prefix=''):
        feature_counter = Counter()
        if version_info[0] >= 3:
            doc = str(doc)
        for category, score in self._analyze(doc).to_dict()[0].items():
            feature_counter[prefix + category] = int(score)
        return feature_counter

    @abstractmethod
    def _get_terms_from_doc(self, doc):
        pass


class FeatsFromTopicModel(FeatsFromTopicModelBase, FeatsFromSpacyDoc):
    def __init__(self,
                 topic_model,
                 use_lemmas=False,
                 entity_types_to_censor=set(),
                 entity_types_to_use=None,
                 tag_types_to_censor=set(),
                 strip_final_period=False,
                 keyword_processor_args={'case_sensitive': False}):
        self._keyword_processor = KeywordProcessor(**keyword_processor_args)
        self._topic_model = topic_model.copy()
        if keyword_processor_args.get('case_sensitive', None) is False:
            for k, v in self._topic_model.items():
                self._topic_model[k] = [e.lower() for e in v]
        for keyphrase in reduce(lambda x, y: set(x) | set(y), self._topic_model.values()):
            self._keyword_processor.add_keyword(keyphrase)
        FeatsFromSpacyDoc.__init__(self, use_lemmas, entity_types_to_censor,
                                   tag_types_to_censor, strip_final_period)
        FeatsFromTopicModelBase.__init__(self, topic_model)

    def get_top_model_term_lists(self):
        return self._topic_model

    def _get_terms_from_doc(self, doc):
        return Counter(self._keyword_processor.extract_keywords(str(doc)))

    def get_feats(self, doc):
        return Counter(self._get_terms_from_doc(str(doc)))


"""
class FeatsFromTopicModel(FeatsFromSpacyDoc, FeatsFromTopicModelBase):
	def __init__(self,
	             topic_model,
	             use_lemmas=False,
	             entity_types_to_censor=set(),
	             tag_types_to_censor=set(),
	             strip_final_period=False,
	             **kwargs):
		'''
		Parameters
		----------
		topic_model : dict
			{topicmodelname: [term1, term2, ....], ...}

		Other parameters from FeatsFromSpacyDoc.__init__
		'''
		check_topic_model_string_format(topic_model)
		self._topic_model = topic_model
		self._lexicon_df = self._get_lexicon_df_from_topic_model(topic_model)
		super(FeatsFromTopicModel, self).__init__(use_lemmas,
		                                          entity_types_to_censor,
		                                          tag_types_to_censor,
		                                          strip_final_period)




	def _get_terms_from_doc(self, doc):
		return Counter(t for t in split(r"(\W)", doc.lower()) if t.strip())

	def has_metadata_term_list(self):
		return True

	def get_top_model_term_lists(self):
		return self._topic_model
"""
